import numpy as np
import scipy.sparse as sprs
from scipy.sparse import csgraph
from openpnm._skgraph.tools import conns_to_am, dict_to_am, dict_to_im
from openpnm._skgraph.tools import istriu, isgtriu
from openpnm._skgraph.tools import get_node_prefix, get_edge_prefix


__all__ = [
    'find_common_edges',
    'filter_by_z',
    'find_connecting_edges',
    'find_neighbor_nodes',
    'find_neighbor_edges',
    'find_connected_nodes',
    'find_complementary_nodes',
    'find_complementary_edges',
    'find_path',
    'find_coordination',
]


def find_complementary_edges(network, inds, asmask=False):
    r"""
    Finds the complementary edges to a given set of inputs

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like
        A list of edge indices for which the complement is sought
    asmask : bool
        If set to ``True`` the result is returned as a boolean mask of the
        correct length with ``True`` values indicate the complements.  The
        default is ``False`` which returns a list of indices instead.

    Returns
    -------
    An array containing indices of the edges that are not part of the input
    list

    """
    edge_prefix = get_edge_prefix(network)
    inds = np.unique(inds)
    N = network[edge_prefix+'.conns'].shape[0]
    mask = np.ones(shape=N, dtype=bool)
    mask[inds] = False
    if asmask:
        return mask
    else:
        return np.arange(N)[mask]


def find_complementary_nodes(network, inds, asmask=False):
    r"""
    Finds the complementary nodes to a given set of inputs

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like (optional)
        A list of indices for which the complement is sought
    asmask : bool
        If set to ``True`` the result is returned as a boolean mask of the
        correct length with ``True`` values indicate the complements. The
        default is ``False`` which returns a list of indices instead.

    Returns
    -------
    An array containing indices of the nodes that are not part of the input
    list

    """
    node_prefix = get_node_prefix(network)
    inds = np.unique(inds)
    N = network[node_prefix+'.coords'].shape[0]
    mask = np.ones(shape=N, dtype=bool)
    mask[inds] = False
    if asmask:
        return mask
    else:
        return np.arange(N)[mask]


def find_connected_nodes(network, inds, flatten=True, logic='or'):
    r"""
    Finds which nodes are connected to a given set of edges

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like
        A list of edges indices whose connected nodes are sought
    flatten : bool (default is ``True``)
        Indicates whether the returned result is a compressed array of all
        neighbors, or a list of lists with each sub-list containing the
        neighbors for each input edge.  Note that an *unflattened* list might
        be slow to generate since it is a Python ``list`` rather than a Numpy
        array.
    logic : str
        Specifies logic to filter the resulting list.  Options are:

        ======= ===============================================================
        logic   Description
        ======= ===============================================================
        'or'    (default) All neighbors of the inputs.  This is also known as
                the 'union' in set theory or 'any' in boolean logic. Both
                keywords are accepted and treated as 'or'.
        'xor'   Only neighbors of one and only one inputs.  This is useful for
                finding neighbors that are not *shared* by any of the input
                nodes. 'exclusive_or' is also accepted.
        'xnor'  Neighbors that are shared by two or more inputs . This is
                equivalent to finding all neighbors with 'or', minus those
                found with 'xor', and is useful for finding neighbors that the
                inputs have in common.  'nxor' is also accepted.
        'and'   Only neighbors shared by all inputs. This is also known as
                'intersection' in set theory and (sometimes) as 'all' in
                boolean logic.  Both keywords are accepted and treated as
                'and'.
        ======= ===============================================================

    Returns
    -------
    An array containing the connected sites, filtered by the given logic.  If
    ``flatten`` is ``False`` then the result is a list of lists containing the
    neighbors of each given input edge.  In this latter case, nodes that
    have been removed by the given logic are indicated by ``nans``, thus the
    array is of type ``float`` and is not suitable for indexing.

    """
    if not isgtriu(network):
        raise Exception("This function is not implemented for directed networks")
    edges = np.array(inds, ndmin=1)
    if len(edges) == 0:  # Short-circuit this function if edges is empty
        return []
    am = dict_to_am(network)
    neighbors = np.hstack((am.row[edges], am.col[edges]))
    if neighbors.size > 0:
        n_sites = np.amax(neighbors)
    if logic in ['or', 'union', 'any']:
        neighbors = np.unique(neighbors)
    elif logic in ['xor', 'exclusive_or']:
        neighbors = np.unique(np.where(np.bincount(neighbors) == 1)[0])
    elif logic in ['xnor', 'nxor']:
        neighbors = np.unique(np.where(np.bincount(neighbors) > 1)[0])
    elif logic in ['and', 'all', 'intersection']:
        temp = np.vstack((am.row[edges], am.col[edges])).T.tolist()
        temp = [set(pair) for pair in temp]
        neighbors = temp[0]
        [neighbors.intersection_update(pair) for pair in temp[1:]]
        neighbors = np.array(list(neighbors), dtype=np.int64, ndmin=1)
    else:
        raise Exception('Specified logic is not implemented')
    if flatten is False:
        if neighbors.size:
            mask = np.zeros(shape=n_sites + 1, dtype=bool)
            mask[neighbors] = True
            temp = np.hstack((am.row[edges], am.col[edges])).astype(np.int64)
            temp[~mask[temp]] = -1
            inds = np.where(temp == -1)[0]
            if len(inds):
                temp = temp.astype(float)
                temp[inds] = np.nan
            temp = np.reshape(temp, newshape=[len(edges), 2], order='F')
            neighbors = temp
        else:
            neighbors = [np.array([], dtype=np.int64) for i in range(len(edges))]
    return neighbors


def find_neighbor_edges(network, inds, flatten=True, logic='or'):
    r"""
    Finds all edges that are connected to the given input nodes

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like (optional)
        A list of node indices whose neighbor edges are sought
    flatten : bool (default is ``True``)
        Indicates whether the returned result is a compressed array of all
        neighbors, or a list of lists with each sub-list containing the
        neighbors for each input node.  Note that an *unflattened* list might
        be slow to generate since it is a Python ``list`` rather than a Numpy
        array.
    logic : str
        Specifies logic to filter the resulting list.  Options are:

        ======= ===============================================================
        logic   Description
        ======= ===============================================================
        'or'    (default) All neighbors of the inputs.  This is also known as
                the 'union' in set theory or 'any' in boolean logic. Both
                keywords are accepted and treated as 'or'.
        'xor'   Only neighbors of one and only one inputs.  This is useful for
                finding neighbors that are not *shared* by any of the input
                nodes. 'exclusive_or' is also accepted.
        'xnor'  Neighbors that are shared by two or more inputs . This is
                equivalent to finding all neighbors with 'or', minus those
                found with 'xor', and is useful for finding neighbors that the
                inputs have in common.  'nxor' is also accepted.
        'and'   Only neighbors shared by all inputs. This is also known as
                'intersection' in set theory and (somtimes) as 'all' in
                boolean logic.  Both keywords are accepted and treated as
                'and'.
        ======= ===============================================================

    Returns
    -------
    An array containing the neighboring edges filtered by the given logic. If
    ``flatten`` is ``False`` then the result is a list of lists containing the
    neighbors of each given input node.

    Notes
    -----
    The ``logic`` options are applied to neighboring edges only, thus it is not
    possible to obtain edges that are part of the global set but not neighbors.
    This is because (a) the list of global edges might be very large, and
    (b) it is not possible to return a list of neighbors for each input site
    if global sites are considered.

    """
    if flatten == False:
        im = dict_to_im(network)
        am = None
    else:
        am = dict_to_am(network)
        im = None
    if im is not None:
        if im.format != 'lil':
            im = im.tolil(copy=False)
        rows = [im.rows[i] for i in np.array(inds, ndmin=1, dtype=np.int64)]
        if len(rows) == 0:
            return []
        neighbors = np.hstack(rows).astype(np.int64)
        n_bonds = int(im.nnz / 2)
        if logic in ['or', 'union', 'any']:
            neighbors = np.unique(neighbors)
        elif logic in ['xor', 'exclusive_or']:
            neighbors = np.unique(np.where(np.bincount(neighbors) == 1)[0])
        elif logic in ['xnor', 'shared']:
            neighbors = np.unique(np.where(np.bincount(neighbors) > 1)[0])
        elif logic in ['and', 'all', 'intersection']:
            neighbors = set(neighbors)
            [neighbors.intersection_update(i) for i in rows]
            neighbors = np.array(list(neighbors), dtype=int, ndmin=1)
        else:
            raise Exception('Specified logic is not implemented')
        if (flatten is False):
            if (neighbors.size > 0):
                mask = np.zeros(shape=n_bonds, dtype=bool)
                mask[neighbors] = True
                for i in range(len(rows)):
                    vals = np.array(rows[i], dtype=np.int64)
                    rows[i] = vals[mask[vals]]
                neighbors = rows
            else:
                neighbors = [np.array([], dtype=np.int64) for i in range(len(inds))]
        return neighbors
    elif am is not None:
        if am.format != 'coo':
            am = am.tocoo(copy=False)
        if flatten is False:
            raise Exception('flatten cannot be used with an adjacency matrix')
        if isgtriu(network):
            am = sprs.triu(am, k=1)
        Ps = np.zeros(am.shape[0], dtype=bool)
        Ps[inds] = True
        conns = np.vstack((am.row, am.col)).T
        if logic in ['or', 'union', 'any']:
            neighbors = np.any(Ps[conns], axis=1)
        elif logic in ['xor', 'exclusive_or']:
            neighbors = np.sum(Ps[conns], axis=1) == 1
        elif logic in ['xnor', 'shared']:
            neighbors = np.all(Ps[conns], axis=1)
        elif logic in ['and', 'all', 'intersection']:
            raise Exception('Specified logic is not implemented')
        else:
            raise Exception('Specified logic is not implemented')
        neighbors = np.where(neighbors)[0]
        return neighbors
    else:
        raise Exception('Either the incidence or the adjacency matrix must be specified')


def find_neighbor_nodes(network, inds, flatten=True, include_input=False,
                        logic='or'):
    r"""
    Finds all nodes that are directly connected to the input nodes

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like
        A list of node indices whose neighbors are sought
    flatten : bool
        If ``True`` (default) the returned result is a compressed array of all
        neighbors, or a list of lists with each sub-list containing the
        neighbors for each input site.  Note that an *unflattened* list might
        be slow to generate since it is a Python ``list`` rather than a Numpy
        array.
    include_input : bool
        If ``False`` (default) the input nodes will be removed from the result.

    logic : str
        Specifies logic to filter the resulting list.  Options are:

        ======= ===============================================================
        logic   Description
        ======= ===============================================================
        'or'    (default) All neighbors of the inputs.  This is also known as
                the 'union' in set theory or 'any' in boolean logic. Both
                keywords are accepted and treated as 'or'.
        'xor'   Only neighbors of one and only one inputs.  This is useful for
                finding neighbors that are not *shared* by any of the input
                nodes. 'exclusive_or' is also accepted.
        'xnor'  Neighbors that are shared by two or more inputs . This is
                equivalent to finding all neighbors with 'or', minus those
                found with 'xor', and is useful for finding neighbors that the
                inputs have in common.  'nxor' is also accepted.
        'and'   Only neighbors shared by all inputs. This is also known as
                'intersection' in set theory and (somtimes) as 'all' in
                boolean logic.  Both keywords are accepted and treated as
                'and'.
        ======= ===============================================================

    Returns
    -------
    nodes : ndarray
        An array containing the neighboring nodes filtered by the given logic.  If
        ``flatten`` is ``False`` then the result is a list of lists containing the
        neighbors of each input site.

    Notes
    -----
    The ``logic`` options are applied to neighboring nodes only, thus it is not
    possible to obtain nodes that are part of the global set but not neighbors.
    This is because the list of global nodes might be very large.

    """
    g = network
    nodes = np.array(inds, ndmin=1)
    # Short-circuit the function if the input list is already empty
    if len(nodes) == 0:
        return []
    am_coo = dict_to_am(g)
    am = am_coo.tolil(copy=False)
    rows = am.rows[nodes].tolist()
    if len(rows) == 0:
        return []
    n_nodes = am.shape[0]
    neighbors = am_coo.col[np.in1d(am_coo.row, nodes)]
    if logic in ['or', 'union', 'any']:
        neighbors = np.unique(neighbors)
    elif logic in ['xor', 'exclusive_or']:
        neighbors = np.unique(np.where(np.bincount(neighbors) == 1)[0])
    elif logic in ['xnor', 'nxor']:
        neighbors = np.unique(np.where(np.bincount(neighbors) > 1)[0])
    elif logic in ['and', 'all', 'intersection']:
        neighbors = set(neighbors)
        [neighbors.intersection_update(i) for i in rows]
        neighbors = np.array(list(neighbors), dtype=np.int64, ndmin=1)
    else:
        raise Exception('Specified logic is not implemented')
    # Deal with removing inputs or not
    mask = np.zeros(shape=n_nodes, dtype=bool)
    mask[neighbors] = True
    if not include_input:
        mask[nodes] = False
    # Finally flatten or not
    if flatten:
        neighbors = np.where(mask)[0]
    else:
        if neighbors.size > 0:
            for i in range(len(rows)):
                vals = np.array(rows[i], dtype=np.int64)
                rows[i] = vals[mask[vals]]
            neighbors = rows
        else:
            neighbors = [np.array([], dtype=int) for i in range(len(nodes))]
    return neighbors


def find_connecting_edges(inds, network=None, am=None):
    r"""
    Finds the edge that connects each pair of given nodes

    Parameters
    ----------
    inds : array_like
        A 2-column vector containing pairs of node indices
    network : dict, optional
        The network dictionary.  Either this or ``am`` must be provided
    am : scipy.sparse matrix, optional
        The adjacency matrix of the network. Must be symmetrical such that if
        nodes *i* and *j* are connected, the matrix contains non-zero values
        at locations (i, j) and (j, i). Either this or ``g`` must be provided.

    Returns
    -------
    edges : ndarray
        An ndarry the same length as P1 (and P2) with each element
        containing the edge number that connects the corresponding nodes,
        or `nan`` if nodes are not connected.

    Notes
    -----
    The adjacency matrix is converted to the ``DOK`` format internally if
    needed, so if this format is already available it should be provided to
    save time.

    """
    nodes = np.array(inds, ndmin=2)
    # Short-circuit function if nodes is an empty list
    if nodes.size == 0:
        return []
    if network is not None:
        edge_prefix = get_edge_prefix(network)
        am = dict_to_am(
            network,
            weights=np.arange(network[edge_prefix+'.conns'].shape[0])
        )
    elif am is not None:
        pass
    else:
        raise Exception('Either g or am must be provided')
    if am.format != 'dok':
        am = am.todok(copy=True)
    z = tuple(zip(nodes[:, 0], nodes[:, 1]))
    neighbors = np.array([am.get(item, np.nan) for item in z])
    return neighbors


def find_common_edges(network, inds_1, inds_2):
    """
    Finds edges shared between two sets of nodes

    Parameters
    ----------
    network : dict
        The network dictionary
    inds_1 : array_like
        A list of indices defining the first set of nodes
    inds_2 : array_like
        A list of indices defining the second set of nodes

    Returns
    -------
    edges : ndarray
        List of edge indices connecting the two given sets of nodes

    """
    if np.intersect1d(inds_1, inds_2).size != 0:
        raise Exception("inds_1 and inds_2 must not share any nodes")
    if not isgtriu(network):
        raise Exception("This function is not implemented for directed graphs")
    edges_1 = find_neighbor_edges(inds=inds_1, network=network, logic="xor")
    edges_2 = find_neighbor_edges(inds=inds_2, network=network, logic="xor")
    return np.intersect1d(edges_1, edges_2)


def filter_by_z(network, inds, z=1):
    r"""
    Filters a list of nodes to those with a given number of neighbors

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like
        A list containing the indices of the nodes to be filtered
    z : int
        The coordination number by which to filter

    Returns
    -------
    inds : array_like
        A list of node indices which satisfy the criteria

    """
    inds = np.array(inds)
    coordination = find_coordination(network)
    hits = coordination == z
    inds = inds[hits[inds]]
    return inds


def find_coordination(network, nodes=None):
    r"""
    Find the coordination number of nodes

    Parameters
    ----------
    network : dict
        The network dictionary
    nodes : array_like, optional
        The nodes for which coordination is sought. If not provided then
        coordination for *all* nodes is returned

    Returns
    -------
    z : ndarray
        An array containing the number of neighbors for each given node

    Notes
    -----
    Supports directed and undirected graphs

    """
    am = dict_to_am(network)
    z = am.getnnz(axis=1)
    if nodes is None:
        return z
    else:
        return z[np.array(nodes)]


def find_path(network, pairs, weights=None):
    r"""
    Find the shortest path between pairs of nodes

    Parameters
    ----------
    network : dict
        The network dictionary
    pairs : array_like
        An N x 2 array containing N pairs of nodes between which the shortest
        path is sought
    weights : ndarray, optional
        The edge weights to use when traversing the path. If not provided
        then 1's will be used.

    Returns
    -------
    paths : dict
        A dictionary containing ``'node_paths'`` and ``'edge_paths'``, each
        containing a list of lists indicating the path between each set of
        nodes given in ``pairs``. An empty list indicates that no path was
        found between a given set of pairs.

    Notes
    -----
    The shortest path is found using Dijkstra's algorithm included in the
    ``scipy.sparse.csgraph`` module

    """
    am = dict_to_am(network)
    if weights is not None:
        am.data = np.ones_like(am.row, dtype=int)
    pairs = np.array(pairs, ndmin=2)
    paths = csgraph.dijkstra(csgraph=am, indices=pairs[:, 0],
                             return_predecessors=True, min_only=False)[1]
    if isgtriu(network):
        am.data = np.hstack(2*[np.arange(am.data.size/2)]).astype(int)
    else:
        am.data = np.arange(am.data.size).astype(int)
    dok = am.todok()
    nodes = []
    edges = []
    for row in range(0, np.shape(pairs)[0]):
        j = pairs[row][1]
        ans = []
        while paths[row][j] > -9999:
            ans.append(j)
            j = paths[row][j]
        if len(ans) > 0:
            ans.append(pairs[row][0])
            ans.reverse()
            nodes.append(np.array(ans, dtype=int))
            keys = [tuple((ans[i], ans[i+1])) for i in range(len(ans)-1)]
            temp = [dok[k] for k in keys]
            edges.append(np.array(temp, dtype=int))
        else:
            nodes.append([])
            edges.append([])
    return {'node_paths': nodes, 'edge_paths': edges}



import numpy as np
from openpnm._skgraph.generators import cubic
from openpnm._skgraph.operations import trim_nodes


def cubic_template(template, spacing=1, connectivity=6,
                   node_prefix='node', edge_prefix='edge'):
    r"""
    Generate a simple cubic lattice matching the shape of the provided tempate

    Parameters
    ----------
    templte : ndarray
        Each ``True`` value will be treated as a vertex while all others
        will be trimmed.
    spacing : array_like or float
        The size of a unit cell in each direction. If an scalar is given it is
        applied in all 3 directions.

    Returns
    -------
    network : dict
        A dictionary containing 'node.coords' and 'edge.conns'

    """
    template = np.atleast_3d(template).astype(bool)
    # Generate a full cubic network
    temp = cubic(shape=template.shape, spacing=spacing,
                 connectivity=connectivity,
                 node_prefix=node_prefix, edge_prefix=edge_prefix)
    # Store some info about template
    coords = np.unravel_index(range(template.size), template.shape)
    coords = np.vstack(coords).T
    Np = coords.shape[0]
    temp[node_prefix+'.template_coords'] = coords
    temp[node_prefix+'.template_indices'] = np.arange(Np)
    # Trim pores not present in template
    temp = trim_nodes(network=temp, inds=~template.flatten())
    return temp


if __name__ == '__main__':
    im = np.ones([50, 50], dtype=bool)
    im[25:, ...] = False
    net = cubic_template(template=im)
    print(net.keys())



r"""
------------
scikit-graph
------------

scikit-graph (or ``skgraph``) is a libary for working with graphs (or
networks). It is built on the following principles:

* Graphs are represented by Python dictionaries filled with numpy ``ndarrays``
* All functionality is in the form of a library of functions that operate on
  the data in the dictionaries with no assumptions about class attributes or
  methods. The data just needs to be formatted and named correctly.
* Nodes and edges have data attributes, such as volume or length, which are
  used in computations
* Networks can be very large (billions of nodes) so good numerical performance
  is critical, to the extent that numpy and related packages can provide

The ``skgraph`` protocol is as follows:

* All the information about a network or graph is stored in a standard
  python ``dict``.
* The ``keys`` must be named following the convention of
  ``'<node_prefix>.<attribute>'`` or ``'<edge_prefix>.<attribute>'``. (The
  prefix 'node' and 'edge' are customizable)
* The dict *must contain* ``'edge.conns'`` and ``'node.coords'``. Everything
  else is optional.
  * ``node.coords`` is an M-by-3 array containg the [x, y, z] locations
  of each node, where M is the number of nodes in the network.
  * ``edge.conns`` is an N-by-2 array of which nodes are connected to
  which, where N is the number of edges in the network.
* node and edge indices are implied by their position in the node or edge
  arrays. For e.g. g['node.coords'][0, :] contains the x,y,z coordinates
  of node 0.
* Undirected graphs are indicated by enforcing the adjacency matrix is in
  upper triangular form. Any entries in the lower triangular portion
  result in a directed graph. For rare edge case where a directed graph
  happens to have all connections in the upper triangle, the only solution
  is to reverse the direction of *all* connections to be in the lower
  triangle. Multigraphs are not supported.

"""
import numpy as _np
from openpnm.utils import PrintableDict as _pdict
from openpnm.utils import SettingsAttr as _Settings


settings = _Settings()
settings.missing_values = {'bool': False,
                           'int': _np.nan,
                           'float': _np.nan,
                           'object': None}


from . import generators
from . import io
from . import operations
from . import queries
from . import simulations
from . import tools
from . import visualization


def info(network):
    r"""
    Prints an overview of the network dictionary
    """
    d = _pdict(network)
    d._key = 'Attribute'
    d._value = 'Description'
    print(d)


from .tools import get_edge_prefix
from .tools import get_node_prefix



import logging
import numpy as np
import scipy.sparse.csgraph as spgr
from openpnm.topotools import is_fully_connected
from openpnm.algorithms import Algorithm
from openpnm.utils import Docorator, TypedSet, Workspace
from openpnm.utils import check_data_health
from openpnm import solvers
from ._solution import SteadyStateSolution, SolutionContainer


__all__ = ['Transport']


docstr = Docorator()
logger = logging.getLogger(__name__)
ws = Workspace()


@docstr.get_sections(base='TransportSettings', sections=['Parameters'])
@docstr.dedent
class TransportSettings:
    """
    Defines the settings for Transport algorithms

    Parameters
    ----------
    %(AlgorithmSettings.parameters)s
    quantity : str
        The name of the physical quantity to be solved for (i.e.
        'pore.concentration')
    conductance : str
        The name of the pore-scale transport conductance values (i.e
        'throat.diffusive_conductance')
    cache : bool
        If ``True``, the ``A`` matrix is cached and rather than getting
        rebuilt.
    variable_props : list of strings
        This list (actually a set) indicates which properties are variable
        and should be updated by the algorithm on each iteration. Note that
        any properties which already depend on ``'quantity'`` will
        automatically be updated.

    """
    phase = ''
    quantity = ''
    conductance = ''
    cache = True
    variable_props = TypedSet()


@docstr.get_sections(base='Transport', sections=['Parameters'])
@docstr.dedent
class Transport(Algorithm):
    """
    This class implements steady-state linear transport calculations.

    Parameters
    ----------
    %(Algorithm.parameters)s

    """

    def __init__(self, phase, name='trans_?', **kwargs):
        super().__init__(name=name, **kwargs)
        self.settings._update(TransportSettings())
        self.settings['phase'] = phase.name
        self['pore.bc.rate'] = np.nan
        self['pore.bc.value'] = np.nan
        self._A = None
        self._b = None
        self._pure_A = None
        self._pure_b = None
        self.soln = {}

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            if key == 'pore.source':
                return {}
            else:
                raise KeyError(key)

    @property
    def x(self):
        """Shortcut to the solution currently stored on the algorithm."""
        return self[self.settings['quantity']]

    @x.setter
    def x(self, value):
        self[self.settings['quantity']] = value

    @docstr.dedent
    def _build_A(self):
        """
        Builds the coefficient matrix based on throat conductance values.

        Notes
        -----
        The conductance to use is specified in stored in the algorithm's
        settings under ``alg.settings['conductance']``.

        """
        gvals = self.settings['conductance']
        if gvals in self.iterative_props:
            self.settings.cache = False
        if not self.settings['cache']:
            self._pure_A = None
        if self._pure_A is None:
            phase = self.project[self.settings.phase]
            g = phase[gvals]
            am = self.network.create_adjacency_matrix(weights=g, fmt='coo')
            self._pure_A = spgr.laplacian(am).astype(float)
        self.A = self._pure_A.copy()

    def _build_b(self):
        """Initializes the RHS vector, b, with zeros."""
        b = np.zeros(self.Np, dtype=float)
        self._pure_b = b
        self.b = self._pure_b.copy()

    @property
    def A(self):
        """The coefficient matrix, A (in Ax = b)"""
        if self._A is None:
            self._build_A()
        return self._A

    @A.setter
    def A(self, value):
        self._A = value

    @property
    def b(self):
        """The right-hand-side (RHS) vector, b (in Ax = b)"""
        if self._b is None:
            self._build_b()
        return self._b

    @b.setter
    def b(self, value):
        self._b = value

    def _apply_BCs(self):
        """Applies specified boundary conditions by modifying A and b."""
        if 'pore.bc.rate' in self.keys():
            # Update b
            ind = np.isfinite(self['pore.bc.rate'])
            self.b[ind] = self['pore.bc.rate'][ind]
        if 'pore.bc.value' in self.keys():
            f = self.A.diagonal().mean()
            # Update b (impose bc values)
            ind = np.isfinite(self['pore.bc.value'])
            self.b[ind] = self['pore.bc.value'][ind] * f
            # Update b (subtract quantities from b to keep A symmetric)
            x_BC = np.zeros_like(self.b)
            x_BC[ind] = self['pore.bc.value'][ind]
            self.b[~ind] -= (self.A * x_BC)[~ind]
            # Update A
            P_bc = self.to_indices(ind)
            mask = np.isin(self.A.row, P_bc) | np.isin(self.A.col, P_bc)
            # Remove entries from A for all BC rows/cols
            self.A.data[mask] = 0
            # Add diagonal entries back into A
            datadiag = self.A.diagonal()
            datadiag[P_bc] = np.ones_like(P_bc, dtype=float) * f
            self.A.setdiag(datadiag)
            self.A.eliminate_zeros()

    def run(self, solver=None, x0=None, verbose=False):
        """
        Builds the A and b matrices, and calls the solver specified in the
        ``settings`` attribute.

        This method stores the solution in the algorithm's ``soln``
        attribute as a ``SolutionContainer`` object. The solution itself
        is stored in the ``x`` attribute of the algorithm as a NumPy array.

        Parameters
        ----------
        x0 : ndarray
            Initial guess of unknown variable

        Returns
        -------
        None

        """
        logger.info('Running Transport')
        if solver is None:
            solver = getattr(solvers, ws.settings.default_solver)()
        # Perform pre-solve validations
        self._validate_settings()
        self._validate_topology_health()
        self._validate_linear_system()
        # Write x0 to algorithm (needed by _update_iterative_props)
        self.x = x0 = np.zeros_like(self.b) if x0 is None else x0.copy()
        self["pore.initial_guess"] = x0
        self._validate_x0()
        # Initialize the solution object
        self.soln = SolutionContainer()
        self.soln[self.settings['quantity']] = SteadyStateSolution(x0)
        self.soln.is_converged = False
        # Build A and b, then solve the system of equations
        self._update_A_and_b()
        self._run_special(solver=solver, x0=x0, verbose=verbose)

    def _run_special(self, solver, x0, w=1.0, verbose=None):
        # Make sure A and b are 'still' well-defined
        self._validate_linear_system()
        # Solve and apply under-relaxation
        x_new, exit_code = solver.solve(A=self.A, b=self.b, x0=x0)
        self.x = w * x_new + (1 - w) * self.x
        # Update A and b using the recent solution otherwise, for iterative
        # algorithms, residual will be incorrectly calculated ~0, since A & b
        # are outdated
        self._update_A_and_b()
        # Update SteadyStateSolution object on algorithm
        self.soln[self.settings['quantity']][:] = self.x
        self.soln.is_converged = not bool(exit_code)

    def _update_A_and_b(self):
        """Builds A and b, and applies specified boundary conditions."""
        self._build_A()
        self._build_b()
        self._apply_BCs()

    def _validate_x0(self):
        """Ensures x0 doesn't contain any nans/infs."""
        x0 = self["pore.initial_guess"]
        if not np.isfinite(x0).all():
            raise Exception("x0 contains inf/nan values")

    def _validate_settings(self):
        if self.settings['quantity'] is None:
            raise Exception("'quantity' hasn't been defined on this algorithm")
        if self.settings['conductance'] is None:
            raise Exception("'conductance' hasn't been defined on this algorithm")
        if self.settings['phase'] is None:
            raise Exception("'phase' hasn't been defined on this algorithm")

    def _validate_topology_health(self):
        """
        Ensures the network is not clustered, and if it is, they're at
        least connected to a boundary condition pore.
        """
        Ps = ~np.isnan(self['pore.bc.rate']) + ~np.isnan(self['pore.bc.value'])
        if not is_fully_connected(network=self.network, pores_BC=Ps):
            msg = ("Your network is clustered, making Ax = b ill-conditioned")
            raise Exception(msg)

    def _validate_linear_system(self):
        """Ensures the linear system Ax = b doesn't contain any nans/infs."""
        if np.isfinite(self.A.data).all() and np.isfinite(self.b).all():
            return
        raise Exception("A or b contains inf/nan values")

    def rate(self, pores=[], throats=[], mode='group'):
        """
        Calculates the net rate of material moving into a given set of
        pores or throats

        Parameters
        ----------
        pores : array_like
            The pores for which the rate should be calculated
        throats : array_like
            The throats through which the rate should be calculated
        mode : str, optional
            Controls how to return the rate. The default value is 'group'.
            Options are:

            ===========  =====================================================
            mode         meaning
            ===========  =====================================================
            'group'      Returns the cumulative rate of material
            'single'     Calculates the rate for each pore individually
            ===========  =====================================================

        Returns
        -------
        If ``pores`` are specified, then the returned values indicate the
        net rate of material exiting the pore or pores.  Thus a positive
        rate indicates material is leaving the pores, and negative values
        mean material is entering.

        If ``throats`` are specified the rate is calculated in the
        direction of the gradient, thus is always positive.

        If ``mode`` is 'single' then the cumulative rate through the given
        pores (or throats) are returned as a vector, if ``mode`` is
        'group' then the individual rates are summed and returned as a
        scalar.

        """
        pores = self._parse_indices(pores)
        throats = self._parse_indices(throats)

        if throats.size > 0 and pores.size > 0:
            raise Exception('Must specify either pores or throats, not both')
        if (throats.size == 0) and (pores.size == 0):
            raise Exception('Must specify either pores or throats')

        network = self.project.network
        phase = self.project[self.settings['phase']]
        g = phase[self.settings['conductance']]

        P12 = network['throat.conns']
        X12 = self.x[P12]
        if g.size == self.Nt:
            g = np.tile(g, (2, 1)).T    # Make conductance an Nt by 2 matrix
        # The next line is critical for rates to be correct
        # We could also do "g.T.flatten()" or "g.flatten('F')"
        g = np.flip(g, axis=1)
        Qt = np.diff(g*X12, axis=1).ravel()

        if throats.size:
            R = np.absolute(Qt[throats])
            if mode == 'group':
                R = np.sum(R)
        elif pores.size:
            Qp = np.zeros((self.Np, ))
            np.add.at(Qp, P12[:, 0], -Qt)
            np.add.at(Qp, P12[:, 1], Qt)
            R = Qp[pores]
            if mode == 'group':
                R = np.sum(R)

        return np.array(R, ndmin=1)

    def clear_value_BCs(self):
        """Clears all value BCs."""
        self.set_BC(pores=None, bctype='value', mode='remove')

    def clear_rate_BCs(self):
        """Clears all rate BCs"""
        self.set_BC(pores=None, bctype='rate', mode='remove')

    def set_value_BC(self, pores=None, values=[], mode='add'):
        """
        Applies constant value boundary conditons to the specified pores.

        These are sometimes referred to as Dirichlet conditions.

        Parameters
        ----------
        pores : array_like
            The pore indices where the condition should be applied
        values : float or array_like
            The value to apply in each pore. If a scalar is supplied
            it is assigne to all locations, and if a vector is applied is
            must be the same size as the indices given in ``pores``.
        mode : str, optional
            Controls how the boundary conditions are applied. The default
            value is 'merge'. For definition of various modes, see the
            docstring for ``set_BC``.

        """
        self.set_BC(pores=pores, bctype='value', bcvalues=values, mode=mode)

    def set_rate_BC(self, pores=None, rates=[], mode='add'):
        """
        Apply constant rate boundary conditons to the specified locations.

        Parameters
        ----------
        pores : array_like
            The pore indices where the condition should be applied
        rates : float or array_like, optional
            The rates to apply in each pore. If a scalar is supplied that
            rate is assigned to all locations, and if a vector is supplied
            it must be the same size as the indices given in ``pores``.
        mode : str, optional
            Controls how the boundary conditions are applied. The default
            value is 'merge'. For definition of various modes, see the
            docstring for ``set_BC``.
        """
        self.set_BC(pores=pores, bctype='rate', bcvalues=rates, mode=mode)



r"""
Phase
=====

This module contains classes generating and handling thermophysical properties

"""


def _fetch_chemical_props(a):
    temp = {}
    k = 1.380649e-23  # Boltzmann constant
    if not hasattr(a, 'zs'):  # If NOT a mixture
        temp['CAS'] = a.CAS
        temp['common_name'] = a.name
        temp['charge'] = a.charge
        temp['formula'] = a.formula
        temp['boiling_temperature'] = a.Tb
        temp['melting_temperature'] = a.Tm
        temp['triple_point_temperature'] = a.Tt
        temp['triple_point_pressure'] = a.Pt
        temp['dipole_moment'] = a.dipole
        temp['LJ_diameter'] = a.molecular_diameter
        temp['LJ_energy'] = a.Stockmayer*k
        # Change temperature to Tb to evaluate surface tension at Tb
        a.T = a.Tb
        temp['surface_tension_Tb'] = a.sigma
        temp['molar_volume_Tb'] = a.Vml_Tb  # Already at Tb
    else:
        temp['formula'] = a.formulas
        temp['mole_fractions'] = a.zs
        temp['CAS'] = None
    temp['molecular_weight'] = a.MW
    temp['critical_temperature'] = a.Tc
    temp['critical_pressure'] = a.Pc
    temp['critical_volume'] = a.Vc
    temp['critical_compressibilty_factor'] = a.Zc
    temp['acentric_factor'] = a.omega
    return temp


def _get_mixture_model_args(
    phase,
    composition='xs',
    args={
        'mus': 'pore.viscosity',
        'MWs': 'param.molecular_weight',
    }
):
    from openpnm.models.phase.misc import mole_to_mass_fraction
    from numpy import vstack
    vals = {}
    if composition in ['ws']:
        temp = vstack(list(mole_to_mass_fraction(phase=phase).values()))[:, 0]
        vals[composition] = temp
    else:
        temp = vstack(list(phase['pore.mole_fraction'].values()))[:, 0]
        vals[composition] = temp
    for item in args.keys():
        temp = vstack(list(phase.get_comp_vals(args[item]).values()))[:, 0]
        vals[item] = temp
    return vals


from ._phase import *
from ._mixture import *
from ._species import *
from ._air import *
from ._water import *
from ._mercury import *



from openpnm.network import Cubic
from openpnm.utils import Docorator
from openpnm.models.collections.geometry import spheres_and_cylinders


docstr = Docorator()


__all__ = ['Cubic']


@docstr.dedent
class Demo(Cubic):
    r"""
    A shortcut for generating a cubic network with geometrical properties
    already added.

    Parameters
    ----------
    %(Network.parameters)s

    """

    def __init__(self, shape=[3, 3, 1], **kwargs):
        super().__init__(shape=shape, **kwargs)
        self.add_model_collection(spheres_and_cylinders)
        self.regenerate_models()



import openpnm.models.physics as mods


basic = {
    'throat.hydraulic_conductance': {
        'model': mods.hydraulic_conductance.generic_hydraulic,
    },
    'throat.diffusive_conductance': {
        'model': mods.diffusive_conductance.generic_diffusive,
    },
    'throat.entry_pressure': {
        'model': mods.capillary_pressure.washburn,
    },
}



import logging
import numpy as np
from openpnm.core import Domain
from openpnm.utils import Workspace, Docorator


docstr = Docorator()
logger = logging.getLogger(__name__)
ws = Workspace()


__all__ = [
    'Phase',
]


@docstr.get_sections(base='PhaseSettings', sections=['Parameters'])
@docstr.dedent
class PhaseSettings:
    r"""
    Parameters
    ----------
    %(BaseSettings.parameters)s
    auto_interpolate : boolean
        If ``True`` then calls to a missing 'throat.<prop>' will automatically
        interpolate 'pore.<prop>', if present, and vice versa.  If ``False``
        a normal ``KeyError`` is raised.
    """
    auto_interpolate = True
    default_domain = 'all'


@docstr.get_sections(base='Phase', sections=['Parameters'])
@docstr.dedent
class Phase(Domain):
    r"""
    This class produces an empty object with no pore-scale models for
    calculating any thermophysical properties. Users must add models and
    specify parameters for all the properties they require.

    Parameters
    ----------
    %(Base.parameters)s

    """

    def __init__(self, network, name='phase_?', **kwargs):
        super().__init__(network=network, name=name, **kwargs)
        self.settings._update(PhaseSettings())
        # Set standard conditions on the phase
        self['pore.all'] = np.ones([network.Np, ], dtype=bool)
        self['throat.all'] = np.ones([network.Nt, ], dtype=bool)
        self['pore.temperature'] = 298.0
        self['pore.pressure'] = 101325.0

    def __setitem__(self, key, value):
        if '@' not in key:
            super().__setitem__(key, value)
        else:
            # Deal with the fact that the label might only exist on the network
            propname, domain = key.split('@')
            element, prop = propname.split('.', 1)
            try:  # Fetch array from self if present
                temp = self[element + '.' + prop]
            except KeyError:  # Otherwise create it
                temp = self._initialize_empty_array_like(value, element)
                self[element + '.' + prop] = temp
            # Insert values into masked locations
            mask = self.project._get_locations(element + '.' + domain)
            temp[mask] = value

    def __getitem__(self, key):
        try:  # If key exists, just get it
            return super().__getitem__(key)
        except KeyError:
            pass

        try:  # Allow look-up from network mostly for label/domain info
            return self.network[key]
        except (KeyError, AttributeError):
            pass

        # Parse the key
        element, prop = key.split('.', 1)
        if '@' in prop:
            prop, domain = prop.split('@')
        else:
            domain = 'all'

        # Get params directly if appropriate
        if element == 'param':
            return self.params[prop]

        # Next get the data arrays, this is the case if @ notation was used
        if element + '.' + prop in self.keys():
            vals = super().__getitem__(element + '.' + prop)
        else:  # If above are not triggered then try to interpolate
            try:
                if self.settings['auto_interpolate']:
                    if (element == 'pore') and ('throat.'+prop not in self.keys()):
                        raise KeyError(key)
                    elif (element == 'throat') and ('pore.'+prop not in self.keys()):
                        raise KeyError(key)
                    vals = self.interpolate_data(element + '.' + prop)
                else:
                    raise KeyError(key)
            except AttributeError:
                raise KeyError(key)

        # Finally get locs
        if domain == 'all':
            locs = np.ones(self._count(element), dtype=bool)
        else:
            locs = self.project._get_locations(element + '.' + domain)

        return vals[locs]



import pickle as pk
from openpnm.network import Network
from openpnm.io import _parse_filename


def network_from_porespy(filename):
    r"""
    Load a network extracted using the PoreSpy package

    Parameters
    ----------
    filename : str or dict
        Can either be a filename pointing to a pickled dictionary, or a
        handle to a dictionary in memory. The second option lets users
        avoid the step of saving the dictionary to a file.

    Returns
    -------
    network : dict
        An OpenPNM network dictionary

    """
    # Parse the filename
    if isinstance(filename, dict):
        net = filename
    else:
        filename = _parse_filename(filename=filename)
        with open(filename, mode='rb') as f:
            net = pk.load(f)

    network = Network()
    network.update(net)

    return network



import numpy as _np
from openpnm.models.geometry import _geodocs


__all__ = [
    "mason_morrow",
    "jenkins_rao",
]


@_geodocs
def mason_morrow(
    network,
    throat_perimeter='throat.perimeter',
    throat_area='throat.cross_sectional_area',
):
    r"""
    Mason and Morrow relate the capillary pressure to the shape factor in a
    similar way to Mortensen but for triangles.

    Parameters
    ----------
    %(network)s
    %(Pt)s
    %(At)s

    Returns
    -------

    References
    ----------
    Mason, G. and Morrow, N.R.. Capillary behavior of a perfectly wetting
    liquid in irregular triangular tubes. Journal of Colloid and Interface
    Science, 141(1), pp.262-274 (1991).

    """
    # Only apply to throats with an area
    ts = network.throats()[network[throat_area] <= 0]
    P = network[throat_perimeter]
    A = network[throat_area]
    value = A/(P**2)
    value[ts] = 1/(4*_np.pi)
    return value


def jenkins_rao(
    network,
    throat_perimeter='throat.perimeter',
    throat_area='throat.cross_sectional_area',
    throat_diameter='throat.indiameter',
):
    r"""
    Jenkins and Rao relate the capillary pressure in an eliptical throat to
    the aspect ratio

    Parameters
    ----------
    %(network)s
    %(Pt)s
    %(At)s
    %(Dt)s

    Returns
    -------

    References
    ----------
    Jenkins, R.G. and Rao, M.B., The effect of elliptical pores on
    mercury porosimetry results. Powder technology, 38(2), pp.177-180. (1984)

    """
    P = network[throat_perimeter]
    A = network[throat_area]
    r = network[throat_diameter]/2
    # Normalized by value for perfect circle
    value = (P/A)/(2/r)
    return value



import numpy as np
import scipy.spatial as sptl
from openpnm._skgraph.tools import vor_to_am, isoutside
from openpnm._skgraph.generators import tools
from openpnm._skgraph.operations import trim_nodes


def voronoi(
    points,
    shape=[1, 1, 1],
    trim=True,
    reflect=False,
    f=1,
    relaxation=0,
    node_prefix='node',
    edge_prefix='edge',
):
    r"""
    Generate a network based on a Voronoi tessellation of base points

    Parameters
    ----------
    points : array_like or int
        Can either be an N-by-3 array of point coordinates which will be used
        directly, or a scalar indicating the number of points to generate.
    shape : array_like
        The size and shape of the domain
    trim : boolean
        If ``True`` (default) then any vertices laying outside the domain
        given by ``shape`` are removed (as are the edges connected to them).
    reflect : boolean, optional (default = ``False``)
        If ``True`` then points are reflected across each face of the domain
        prior to performing the tessellation. Enabling this behavior creates
        flat faces on all sizes of the domain.
    f : float
        The fraction of points which should be reflected.  The default is 1 which
        reflects all the points in the domain, but this can lead to a lot of
        unnecessary points, so setting to 0.1 or 0.2 helps speed, but risks that
        the tessellation may not have smooth faces if not enough points are
        reflected.
    relaxation : int, optional (default = 0)
        The number of iterations to use for relaxing the base points. This is
        sometimes called `Lloyd's algorithm
        <https://en.wikipedia.org/wiki/Lloyd%27s_algorithm>`_. This function computes
        the new base points as the simple average of the Voronoi vertices instead
        of rigorously finding the center of mass, which is quite time consuming.
        To use the rigorous method, call the ``lloyd_relaxation`` function manually
        to obtain relaxed points, then pass the points directly to this funcion.
        The results are quite stable after only a few iterations.

    Returns
    -------
    network : dict
        A dictionary containing node coordinates and edge connections
    vor : Voronoi tessellation object
        The Voronoi tessellation object produced by ``scipy.spatial.Voronoi``

    """
    points = tools.parse_points(points=points, shape=shape, reflect=reflect, f=f)
    mask = ~np.all(points == 0, axis=0)
    # Perform tessellation
    vor = sptl.Voronoi(points=points[:, mask])
    for _ in range(relaxation):
        points = tools.lloyd_relaxation(vor, mode='rigorous')
        # Reparse points
        d = {}
        d[node_prefix+'.coords'] = points
        keep = ~isoutside(network=d, shape=shape)
        points = points[keep]
        points = tools.parse_points(points=points, shape=shape, reflect=reflect)
        vor = sptl.Voronoi(points=points[:, mask])
    # Convert to adjecency matrix
    coo = vor_to_am(vor)
    # Write values to dictionary
    d = {}
    conns = np.vstack((coo.row, coo.col)).T
    d[edge_prefix+'.conns'] = conns
    # Convert coords to 3D if necessary
    # Rounding is crucial since some voronoi verts endup outside domain
    pts = np.around(vor.vertices, decimals=10)
    if mask.sum() < 3:
        coords = np.zeros([pts.shape[0], 3], dtype=float)
        coords[:, mask] = pts
    else:
        coords = pts
    d[node_prefix+'.coords'] = coords
    if trim:
        hits = isoutside(d, shape=shape)
        d = trim_nodes(d, hits)
    return d, vor



import numpy as np


__all__ = [
    'rotate_coords',
    'shear_coords',
    'generate_points_on_sphere',
    'generate_points_in_disk',
    'generate_points_on_circle',
    'cart2sph',
    'sph2cart',
    'cart2cyl',
    'cyl2cart',
]


def rotate_coords(coords, a=0, b=0, c=0, R=None):
    r"""
    Rotates coordinates a given amount about each axis

    Parameters
    ----------
    coords : ndarray
        The site coordinates to be transformed.  ``coords`` must be in 3D,
        but a 2D network can be represented by putting 0's in the missing
        dimension.
    a, b, c : scalar, optional
        The amount in degrees to rotate about the x, y, and z-axis,
        respectively.
    R : array_like, optional
        Rotation matrix.  Must be a 3-by-3 matrix since coordinates are
        always in 3D.  If this is given then `a`, `b`, and `c` are ignored.

    Returns
    -------
    coords : ndarray
        A copy of the given ``coords`` is made and returned to the rotation
        does not occur *in place*.

    See Also
    --------
    shear_coords

    """
    coords = np.copy(coords)
    if R is None:
        if a:
            R = np.array([[1, 0, 0],
                          [0, np.cos(np.deg2rad(a)), -np.sin(np.deg2rad(a))],
                          [0, np.sin(np.deg2rad(a)), np.cos(np.deg2rad(a))]])
            coords = np.tensordot(coords, R, axes=(1, 1))
        if b:
            R = np.array([[np.cos(np.deg2rad(b)), 0, -np.sin(np.deg2rad(b))],
                          [0, 1, 0],
                          [np.sin(np.deg2rad(b)), 0, np.cos(np.deg2rad(b))]])
            coords = np.tensordot(coords, R, axes=(1, 1))
        if c:
            R = np.array([[np.cos(np.deg2rad(c)), -np.sin(np.deg2rad(c)), 0],
                          [np.sin(np.deg2rad(c)), np.cos(np.deg2rad(c)), 0],
                          [0, 0, 1]])
            coords = np.tensordot(coords, R, axes=(1, 1))
    else:
        coords = np.tensordot(coords, R, axes=(1, 1))
    return coords


def shear_coords(coords, ay=0, az=0, bx=0, bz=0, cx=0, cy=0, S=None):
    r"""
    Shears the coordinates a given amount about along axis

    Parameters
    ----------
    coords : ndarray
        The coordinates to be transformed
    ay : scalar
        The factor by which to shear along the x-axis as a function of y
    az : scalar
        The factor by which to shear along the x-axis as a function of z
    bx : scalar
        The factor by which to shear along the y-axis as a function of x
    bz : scalar
        The factor by which to shear along the y-axis as a function of z
    cx : scalar
        The factor by which to shear along the z-axis  as a function of x
    cy : scalar
        The factor by which to shear along the z-axis as a function of y
    S : array_like
        The shear matrix.  Must be a 3-by-3 matrix since pore coordinates are
        always in 3D.  If this is given then the other individual arguments
        are ignored.

    Returns
    -------
    coords : ndarray
        The sheared coordinates.  A copy of the supplied coordinates is made
        so that the operation is not performed *in place*.

    See Also
    --------
    rotate_coords

    Notes
    -----
    The shear along the i *th*-axis is given as i\* = i + aj.  This means
    the new i coordinate is the old one plus some linear factor *a* in the
    j *th* direction.

    The values of ``a``, ``b``, and ``c`` are essentially the inverse of the
    slope to be formed by the neighboring layers of sheared pores.  A value of
    0 means no shear, and neighboring points are stacked directly on top of
    each other; a value of 1 means they form a 45 degree diagonal, and so on.

    If ``S`` is given, then is should be of the form:

    ::

        S = [[1 , ay, az],
             [bx, 1 , bz],
             [cx, cy, 1 ]]

        where any of the off-diagonal components can be 0 meaning no shear

    """
    coords = np.copy(coords)
    if S is None:
        S = np.array([[1, ay, az],
                      [bx, 1, bz],
                      [cx, cy, 1]])
    coords = (S@coords.T).T
    return coords


def generate_points_on_sphere(n=100, r=1):
    r"""
    Generates approximately equispaced points on the surface of a sphere

    Parameters
    ----------
    n : int or [int, int]
        If a single ``int`` is provided then this number of points will be
        generated using the Fibonacci method to make them approximately
        equally spaced.  If a list of 2 ``int``s is given, they are interpreted
        as the number of meridians and parallels to divide the sphere with.
    r : scalar
        The radius of the sphere on which the points should lie

    Returns
    -------
    coords : ndarray
        An array of x, y, z coordinates for the sphere which will be centered
        on [0, 0, 0]

    """
    if isinstance(n, int):
        i = np.arange(n)
        phi = np.pi * (3 - np.sqrt(5))  # golden angle in radians
        y = 1 - (i / float(n - 1))*2  # y goes from 1 to -1
        radius = np.sqrt(1 - y*y)  # radius at y
        theta = phi * i  # golden angle increment
        x = np.cos(theta) * radius
        z = np.sin(theta) * radius
        # Convert to spherical coords
        r_, q, p = cart2sph(x, y, z)
        # Scale the radius then convert back to cartesian
        X, Y, Z = sph2cart(r=r*r_, theta=q, phi=p)
        coords = np.vstack((X, Y, Z)).T
    else:
        nlat = n[0]
        nlon = n[1]
        lat = []
        lon = []
        for i in range(0, 360, max(1, int(360/nlon))):
            for j in range(max(1, int(180/nlat)), 180, max(1, int(180/nlat))):
                lon.append(i)
                lat.append(j)
        lat.extend([0, 180])
        lon.extend([0, 0])
        theta = np.deg2rad(lon) - np.pi
        phi = np.deg2rad(lat) - np.pi/2
        X, Y, Z = sph2cart(phi=phi, theta=theta, r=r)
        coords = np.vstack((X, Y, Z)).T
    return coords


def generate_points_in_disk(n=100, r=1):
    r"""
    Generates approximately equally spaced points inside a disk

    Parameters
    ----------
    n : int
        The number of points to generate
    r : scalar
        The radius of the disk

    Returns
    -------
    coords : ndarray
        An ``n by 2`` array of x, y points (in cartesian coordinates)
    """
    indices = np.arange(0, n, dtype=float) + 0.5
    r = np.sqrt(indices/n)
    theta = np.pi*(1 + 5**0.5)*indices
    x = r*np.cos(theta)
    y = r*np.sin(theta)
    return np.vstack((x, y)).T


def generate_points_on_circle(n=100, r=1):
    r"""
    Generates equally spaced points on a circle

    Parameters
    ----------
    n : int
        The number of points to generate
    r : scalar
        The radius of the disk

    Returns
    -------
    coords : ndarray
        An ``n by 2`` array of x, y points (in cartesian coordinates)
    """
    theta = np.linspace(0, 2*np.pi, n, endpoint=False)
    r = np.ones_like(theta)*r
    z = np.zeros_like(theta)
    x, y, z = cyl2cart(r=r, theta=theta, z=z)
    return np.vstack((x, y)).T


def cart2sph(x, y, z):
    r"""
    Converts cartesian to spherical coordinates

    Parameters
    ----------
    x, y, z : array_like
        Arrays containing the x, y and z coordinates to be converted

    Returns
    -------
    r, theta, phi : ndarrays
        Three arrays containing the spherical coordinate of each given point

    Notes
    -----
    Surprizingly (and annoyingly) this is not built into numpy, for reasons
    discussed `here <https://github.com/numpy/numpy/issues/5228>`_.
    """
    hxy = np.hypot(x, y)
    r = np.hypot(hxy, z)
    phi = np.arctan2(z, hxy)
    theta = np.arctan2(y, x)
    return r, theta, phi


def sph2cart(r, theta, phi):
    r"""
    Converts spherical to cartesian coordinates

    Parameters
    ----------
    r, theta, phi : array_like
        Arrays containing the r, theta and phi coordinates to be transformed

    Returns
    -------
    x, y, z : ndarrays
        Three arrays containing the cartesian coordinates of the given points

    Notes
    -----
    Surprizingly (and annoyingly) this is not built into numpy, for reasons
    discussed `here <https://github.com/numpy/numpy/issues/5228>`_.
    """
    rcos_theta = r * np.cos(phi)
    x = rcos_theta * np.cos(theta)
    y = rcos_theta * np.sin(theta)
    z = r * np.sin(phi)
    return x, y, z


def cart2cyl(x, y, z=0):
    r"""
    Converts cartesian to cylindrical coordinates

    Parameters
    ----------
    x, y, z : array_like
        Arrays containing the cartesian coordinates to be transformed. Note
        that ``z`` is optional since it is unchanged during this operation.
        If ``z`` is omitted it is assumed to be 0, or equivalent to polar
        coordinates.

    Returns
    -------
    r, theta, z : ndarrays
        Three arrays containing the cylindrical coordinates of the given points

    Notes
    -----
    Surprizingly (and annoyingly) this is not built into numpy, for reasons
    discussed `here <https://github.com/numpy/numpy/issues/5228>`_.
    """
    theta = np.arctan2(y, x)
    r = np.hypot(x, y)
    if isinstance(z, (int, float)):
        z = np.ones_like(r)*z
    return r, theta, z


def cyl2cart(r, theta, z=0):
    r"""
    Converts cylindrical to cartesian coordinates

    Parameters
    ----------
    r, theta, z : array_like
        Arrays containing the cylindrical coordinates to be transformed. Note
        that ``z`` is optional since it is unchanged during this operation.
        If ``z`` is omitted it is assumed to be 0, or equivalent to polar
        coordinates.

    Returns
    -------
    x, y, z : ndarrays
        Three arrays containing the cartesian coordinates of the given points

    Notes
    -----
    Surprizingly (and annoyingly) this is not built into numpy, for reasons
    discussed `here <https://github.com/numpy/numpy/issues/5228>`_.
    """
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    if isinstance(z, (int, float)):
        z = np.ones_like(x)*z
    return x, y, z



"""
Health Checks
-------------

"""
import logging
import numpy as np
import pandas as pd
logger = logging.getLogger(__name__)

__all__ = [
    'bidirectional_throats',
    'cluster_number',
    'cluster_size',
    'count_coincident_pores',
    'duplicate_throats',
    'find_coincident_pores',
    'isolated_pores',
    'reversed_throats',
    'looped_throats',
    'headless_throats',
]


def cluster_number(network):
    r"""
    Assign a cluster number to each pore
    """
    from scipy.sparse import csgraph as csg
    am = network.create_adjacency_matrix(fmt='coo', triu=True)
    N, Cs = csg.connected_components(am, directed=False)
    return Cs


def cluster_size(network, cluster=None):
    r"""
    Find the size of the cluster to which each pore belongs

    Parameters
    ----------
    network : dict
        The Network
    cluster : str, optional
        Dict key pointing to the array containing the cluster number of each
        pore.  If not provided then it will be calculated.

    Returns
    -------
    cluster_size : ndarray
        An Np-long array containing the size of the cluster to which each pore
        belongs

    """
    if cluster is None:
        from scipy.sparse import csgraph as csg
        am = network.create_adjacency_matrix(fmt='coo', triu=True)
        N, cluster_num = csg.connected_components(am, directed=False)
    else:
        cluster_num = network[cluster]
    Cs, ind, N = np.unique(cluster_num, return_inverse=True, return_counts=True)
    values = N[ind]
    return values


def isolated_pores(network):
    r"""
    Find which pores, if any, are not connected to a throat
    """
    values = np.ones(network.Np, dtype=bool)
    hits = np.unique(network.conns)
    if np.any(hits >= network.Np):
        logger.warning("Some throats point to non-existent pores")
        hits = hits[hits < network.Np]
    values[hits] = False
    return values


def reversed_throats(network):
    r"""
    Find any throat connections that are pointing from j -> i where j > i
    """
    hits = network.conns[:, 0] > network.conns[:, 1]
    return hits


def looped_throats(network):
    r"""
    Find any throats that are connected to the same pore on both ends
    """
    hits = network.conns[:, 0] == network.conns[:, 1]
    return hits


def headless_throats(network):
    r"""
    Find any throats that point to a non-existent pore
    """
    hits = np.any(network.conns > (network.Np - 1), axis=1)
    return hits


def duplicate_throats(network):
    r"""
    Find repeat occurrences of throat connections
    """
    return pd.DataFrame(network.conns).duplicated().to_numpy()


def count_coincident_pores(network, thresh=1e-6):
    r"""
    Count number of pores that are spatially coincident with other pores

    Parameters
    ----------
    network : dict
        The Network
    thresh : float
        The distance below which two pores are considered spatially coincident

    Returns
    -------
    count : ndarray
        A numpy array of Np length containing the number of coincident pores
    """
    # This needs to be a bit complicated because it cannot be assumed
    # the coincident pores are topologically connected
    import scipy.spatial as sptl
    coords = network.coords
    tree = sptl.KDTree(coords)
    hits = tree.query_pairs(r=thresh)
    arr = np.array(list(hits)).flatten()
    v, n = np.unique(arr, return_counts=True)
    values = np.zeros(network.Np, dtype=int)
    values[v.astype(int)] = n
    return values


def find_coincident_pores(network, thresh=1e-6):
    r"""
    Find the indices of coincident pores

    Parameters
    ----------
    network : dict
        The Network
    thresh : float
        The distance below which two pores are considered spatially coincident

    Returns
    -------
    indices : list of lists
        One row corresponding to each pore, with each row listing the indices
        of any coincident pores.  An empty list means no pores were found
        within a distance of ``thresh``.

    """
    # This needs to be a bit complicated because it cannot be assumed
    # the coincident pores are topologically connected
    import scipy.spatial as sptl
    coords = network['pore.coords']
    tree = sptl.KDTree(coords)
    a = tree.sparse_distance_matrix(tree, max_distance=thresh,
                                    output_type='coo_matrix')
    a.data += 1.0
    a.setdiag(0)
    a.eliminate_zeros()
    a.data -= 1.0
    a = a.tolil()
    return a.rows


def bidirectional_throats(network):
    biTs = network['throat.conns'][:, 0] > network['throat.conns'][:, 1]
    return biTs



import logging
import numpy as np
import scipy.stats as spts
from openpnm.models import _doctxt

logger = logging.getLogger(__name__)


__all__ = [
    'random',
    'weibull',
    'normal',
    'generic_distribution',
    'match_histogram',
]


@_doctxt
def weibull(network, seeds, shape, scale, loc):
    r"""
    Produces values from a Weibull distribution given a set of random numbers.

    Parameters
    ----------
    %(network)s
    seeds : str (dict key)
        %(dict_blurb) seed
    shape : float
        Controls the width or skewness of the distribution. For more
        information on the effect of this parameter refer to the
        corresponding `scipy.stats function
        <https://docs.scipy.org/doc/scipy/reference/stats.html>`_.
    scale : float
        Controls the width of the distribution. For more information on
        the effect of this parameter refer to the corresponding
        scipy.stats function.
    loc : float
            Specifies the central value of ???

    Returns
    -------
    values : ndndarray
        A numpy ndarray containing values following the distribution

    Examples
    --------
    The following code illustrates the inner workings of this function,
    which uses the 'weibull_min' method of the scipy.stats module.  This can
    be used to find suitable values of 'shape', 'scale'` and 'loc'.  Note that
    'shape' is represented by 'c' in the actual function call.

    .. plot::

       import numpy
       import scipy.stats
       import matplotlib.pyplot as plt

       func = scipy.stats.weibull_min(c=1.5, scale=0.0001, loc=0)
       plt.hist(func.ppf(q=numpy.random.rand(10000)), bins=50)

       plt.show()

    """
    seeds = network[seeds]
    value = spts.weibull_min.ppf(q=seeds, c=shape, scale=scale, loc=loc)
    return value


@_doctxt
def normal(network, seeds, mean=None, stddev=None, scale=None, loc=None):
    r"""
    Produces values from a Weibull distribution given a set of random numbers.

    Parameters
    ----------
    %(network)s
    seeds : str (dict key)
        %(dict_blurb) seed
    mean : float
        The mean value of the distribution.  This is referred to as the
        ``loc`` in the scipy.stats function, and this key word is also
        accepted.
    stddev : float
        The standard deviation of the distribution.  This is referred to as
        the  ``scale`` in the scipy.stats function, and this key word is also
        accepted.

    Returns
    -------

    Examples
    --------
    The following code illustrates the inner workings of this function,
    which uses the 'norm' method of the scipy.stats module.  This can
    be used to find suitable values of 'scale' and 'loc'.

    .. plot::

       import numpy
       import scipy.stats
       import matplotlib.pyplot as plt

       func = scipy.stats.norm(scale=.0001, loc=0.001)
       fig = plt.hist(func.ppf(q=numpy.random.rand(10000)), bins=50)

       plt.show()

    """
    scale = stddev if stddev is not None else scale
    loc = mean if mean is not None else loc
    seeds = network[seeds]
    value = spts.norm.ppf(q=seeds, scale=scale, loc=loc)
    return value


@_doctxt
def generic_distribution(network, seeds, func, **kwargs):
    r"""
    Accepts an object from the Scipy.stats submodule and returns
    values from the distribution for the given seeds



    Parameters
    ----------
    %(network)s
    seeds : str (dict key)
        %(dict_blurb) seed
    func : object
        An object from the scipy.stats library. Can be a 'frozen' object, where all
        the parameters were specified upon creation, or a handle to an unintialized
        object. In the latter case the parameters for the distribution should be
        provided as keyword arguments.

    Returns
    -------
    values : ndarray
        An ndarray of either Np or Nt length, with values taken from the supplied
        distribution.

    Examples
    --------
    The following code illustrates the process of obtaining a 'frozen' Scipy
    stats object and adding it as a model:

    .. plot::

       import numpy
       import scipy.stats
       import openpnm as op
       import matplotlib.pyplot as plt

       pn = op.network.Cubic(shape=[3, 3, 3])
       pn.add_model(propname='pore.seed',
                    model=op.models.geometry.pore_seed.random)

       # Now retrieve the stats distribution and add to ``geo`` as a model
       stats_obj = scipy.stats.weibull_min(c=2, scale=.0001, loc=0)
       pn.add_model(propname='pore.size',
                    model=op.models.geometry.pore_size.generic_distribution,
                    seeds='pore.seed',
                    func=stats_obj)

       plt.hist(stats_obj.ppf(q=numpy.random.rand(1000)), bins=50)

       plt.show()

    """
    if hasattr(func, 'freeze'):
        func = func.freeze(**kwargs)
    seeds = network[seeds]
    value = func.ppf(seeds)
    return value


@_doctxt
def random(network, element, seed=None, num_range=[0, 1]):
    r"""
    Create an array of random numbers of a specified size.

    Parameters
    ----------
    %(network)s
    seed : int
        The starting seed value to sent to numpy's random number generator.
        A value of ``None`` means a different distribution is returned each
        time the model is (re)run.
    num_range : list
        A two element list indicating the low and high end of the returned
        numbers.  The default is ``[0, 1]``, but a value of ``[0.1, 0.9]``
        may be useful if these values are to be used in subsequent
        distributions to prevent generating extreme values in the tails.

    Returns
    -------

    """
    range_size = num_range[1] - num_range[0]
    range_min = num_range[0]
    if seed is not None:
        np.random.seed(seed)
    value = np.random.rand(network._count(element),)
    value = value*range_size + range_min
    return value


@_doctxt
def match_histogram(network, bin_centers, bin_heights, element='pore'):
    r"""
    Generate values corresponding to a given histogram

    Parameters
    ----------
    %(network)s
    bin_centers : array_like
        The x-axis of the histogram, such as pore sizes.
    bin_heights : array_like
        The y-axis of the histogram, such as the number of pores of each size.
    element : str
        Controls how many values to generate. Can either be 'pore' or 'throat'.

    Returns
    -------
    values : ndarray
        Values corresponding to ``bin_centers`` generated in proportion to the
        respective ``bin_heights``.

    """
    N = network._count(element)
    h = np.cumsum(bin_heights)
    b = np.digitize(np.random.rand(N)*np.amax(h), bins=h)
    vals = np.array(bin_centers)[b]
    return vals



import logging
from openpnm.algorithms import TransientReactiveTransport, FickianDiffusion


__all__ = ['TransientFickianDiffusion']


logger = logging.getLogger(__name__)


class TransientFickianDiffusion(TransientReactiveTransport, FickianDiffusion):
    r"""
    A class to simulate transient diffusion with reactions
    """

    def __init__(self, name='trans_fick_?', **kwargs):
        super().__init__(name=name, **kwargs)



r"""
Main classes of OpenPNM
=======================

This module contains the main classes from which all other major objects derive.

The Base class
--------------

The ``Base`` class is a ``dict`` that has added methods for indexing the pores
and throats, applying labels, and managing the stored data. All OpenPNM
object inherit from ``Base`` so possess these methods.

"""

from ._models import *
from ._mixins import *
from ._base2 import *



r"""
Collection of pore-scale models
===============================

"""

from ._doctxt import *
from . import misc
from . import network
from . import geometry
from . import phase
from . import physics
from . import collections



import numpy as np
import scipy.sparse as sprs
from scipy.spatial import KDTree, distance_matrix
from scipy.spatial import ConvexHull
from scipy.spatial import Delaunay
from scipy.sparse import csgraph
from openpnm._skgraph.tools import generate_points_on_sphere
from openpnm._skgraph.tools import generate_points_on_circle
from openpnm._skgraph.tools import cart2sph, sph2cart, cart2cyl, cyl2cart


# Once a function has been stripped of all its OpenPNM dependent code it
# can be added to this list of functions to import
__all__ = [
    'get_edge_prefix',
    'get_node_prefix',
    'change_prefix',
    'isoutside',
    'dimensionality',
    'find_surface_nodes',
    'find_surface_nodes_cubic',
    'find_coincident_nodes',
    'internode_distance',
    'tri_to_am',
    'vor_to_am',
    'conns_to_am',
    'dict_to_am',
    'dict_to_im',
    'istriu',
    'istril',
    'isgtriu',
    'istriangular',
    'issymmetric',
    'get_cubic_shape',
    'get_cubic_spacing',
    'is_fully_connected',
    'get_domain_length',
    'get_domain_area',
]


def get_edge_prefix(network):
    r"""
    Determines the prefix used for edge arrays from ``<edge_prefix>.conns``

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    edge_prefix : str
        The value of ``<edge_prefix>`` used in ``g``.  This is found by
        scanning ``g.keys()`` until an array ending in ``'.conns'`` is found,
        then returning the prefix.

    Notes
    -----
    This process is surprizingly fast, on the order of nano seconds, so this
    overhead is worth it for the flexibility it provides in array naming.
    However, since all ``dict`` are now sorted in Python, it may be helpful
    to ensure the ``'conns'`` array is near the beginning of the list.
    """
    for item in network.keys():
        if item.endswith('.conns'):
            return item.split('.')[0]


def get_node_prefix(network):
    r"""
    Determines the prefix used for node arrays from ``<edge_prefix>.coords``

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    node_prefix : str
        The value of ``<node_prefix>`` used in ``g``.  This is found by
        scanning ``g.keys()`` until an array ending in ``'.coords'`` is found,
        then returning the prefix.

    Notes
    -----
    This process is surprizingly fast, on the order of nano seconds, so this
    overhead is worth it for the flexibility it provides in array naming.
    However, since all ``dict`` are now sorted in Python, it may be helpful
    to ensure the ``'conns'`` array is near the beginning of the list.
    """
    for item in network.keys():
        if item.endswith('.coords'):
            return item.split('.')[0]


def change_prefix(network, old_prefix, new_prefix):
    r"""
    Changes the prefix used when generating the graph

    Parameters
    ----------
    network : dict
        The network graph
    old_prefix : str
        The current prefix to change, can either be a node or an edge prefix
    new_prefix : str
        The prefix to use instead

    Returns
    -------
    network : dict
        The graph dictionary will arrays assigned to new keys
    """
    for key in list(network.keys()):
        if key.startswith(old_prefix):
            temp = key.split('.', 1)[1]
            network[new_prefix + '.' + temp] = network.pop(key)
    return network


def isoutside(network, shape, rtol=0.0):
    r"""
    Identifies sites that lie outside the specified shape

    Parameters
    ----------
    network : dict
        The network dictionary. For convenience it is also permissible to just
        supply an N-by-D array of coordinates.
    shape : array_like
        The shape of the domain beyond which points are considered "outside".
        The argument is treated as follows:

        ========== ============================================================
        shape      Interpretation
        ========== ============================================================
        [x, y, z]  A 3D cubic domain of dimension x, y and z with the origin at
                   [0, 0, 0].
        [x, y, 0]  A 2D square domain of size x by y with the origin at
                   [0, 0]
        [r, z]     A 3D cylindrical domain of radius r and height z whose
                   central axis starts at [0, 0, 0]
        [r, 0]     A 2D circular domain of radius r centered on [0, 0] and
                   extending upwards
        [r]        A 3D spherical domain of radius r centered on [0, 0, 0]
        ========== ============================================================

    rtol : scalar or array_like, optional
        Controls how far a node must be from the domain boundary to be
        considered outside. It is applied as a fraction of the domain size as
        ``x[i] > (shape[0] + shape[0]*threshold)`` or
        ``y[i] < (0 - shape[1]*threshold)``.  Discrete threshold values
        can be given for each axis by supplying a list the same size as
        ``shape``.

    Returns
    -------
    mask : boolean ndarray
        A boolean array with ``True`` values indicating nodes that lie outside
        the domain.

    Notes
    -----
    If the domain is 2D, either a circle or a square, then the z-dimension
    of ``shape`` should be set to 0.

    """
    try:
        node_prefix = get_node_prefix(network)
        coords = network[node_prefix+'.coords']
    except AttributeError:
        coords = network
    shape = np.array(shape, dtype=float)
    if np.isscalar(rtol):
        tolerance = np.array([rtol]*len(shape))
    else:
        tolerance = np.array(rtol)
    # Label external pores for trimming below
    if len(shape) == 1:  # Spherical
        # Find external points
        R, Q, P = cart2sph(*coords.T)
        thresh = tolerance[0]*shape[0]
        Ps = R > (shape[0] + thresh)
    elif len(shape) == 2:  # Cylindrical
        # Find external pores outside radius
        R, Q, Z = cart2cyl(*coords.T)
        thresh = tolerance[0]*shape[0]
        Ps = R > shape[0]*(1 + thresh)
        # Find external pores above and below cylinder
        if shape[1] > 0:
            thresh = tolerance[1]*shape[1]
            Ps = Ps + (coords[:, 2] > (shape[1] + thresh))
            Ps = Ps + (coords[:, 2] < (0 - thresh))
        else:
            pass
    elif len(shape) == 3:  # Rectilinear
        thresh = tolerance*shape
        Ps1 = np.any(coords > (shape + thresh), axis=1)
        Ps2 = np.any(coords < (0 - thresh), axis=1)
        Ps = Ps1 + Ps2
    return Ps


def dimensionality(network, cache=True):
    r"""
    Checks the dimensionality of the network

    Parameters
    ----------
    network : dict
        The network dictionary
    cache : boolean, optional (default is True)
        If ``False`` then the dimensionality is recalculated even if it has
        already been calculated and stored in the graph dictionary.

    Returns
    -------
    dims : list
        A  3-by-1 array containing ``True`` for each axis that contains
        multiple values, indicating that the pores are spatially distributed
        in that dimension.

    """
    if cache:
        try:
            return network.params["dimensionality"]
        # union of KeyErroa and AttributeError
        except (KeyError, AttributeError):
            pass
    n = get_node_prefix(network)
    coords = network[n+'.coords']
    eps = np.finfo(float).resolution
    dims_unique = \
        [not np.allclose(xk, xk.mean(), atol=0, rtol=eps) for xk in coords.T]
    if cache:
        network["params.dimensionality"] = np.array(dims_unique)
    return np.array(dims_unique)


def find_surface_nodes_cubic(network):
    r"""
    Identifies nodes on the outer surface of the domain assuming a cubic domain
    to save time

    Parameters
    ----------
    network : dict
        The graph dictionary

    Returns
    -------
    mask : ndarray
        A boolean array of ``True`` values indicating which nodes were found
        on the surfaces.
    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    hits = np.zeros(coords.shape[0], dtype=bool)
    dims = dimensionality(network)
    for d in range(3):
        if dims[d]:
            hi = np.where(coords[:, d] == coords[:, d].max())[0]
            lo = np.where(coords[:, d] == coords[:, d].min())[0]
            hits[hi] = True
            hits[lo] = True
    return hits


def find_coincident_nodes(network):
    r"""
    Finds nodes with identical coordinates

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    duplicates : list of ndarrays
        A list with each sublist indicating the indices of nodes that share
        a common set of coordinates

    Notes
    -----
    This function works by computing a ``hash`` of the coordinates then finding
    all nodes with equivalent hash values. Hashes are supposed to be unique
    but they occassionally "collide", meaning nodes may be identified as
    coincident that are not.
    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    hashed = [hash(row.tobytes()) for row in coords]
    uniq, counts = np.unique(hashed, return_counts=True)
    hits = np.where(counts > 1)[0]
    dupes = []
    for item in hits:
        dupes.append(np.where(hashed == uniq[item])[0])
    return dupes


def find_surface_nodes(network):
    r"""
    Identifies nodes on the outer surface of the domain using a Delaunay
    tessellation

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    mask : ndarray
        A boolean array of ``True`` values indicating which nodes were found
        on the surfaces.

    Notes
    -----
    This function generates points around the domain the performs a Delaunay
    tesselation between these points and the network nodes.  Any network
    nodes which are connected to a generated points is considered a surface
    node.

    """
    node_prefix = get_node_prefix(network)
    coords = np.copy(network[node_prefix+'.coords'])
    shift = np.mean(coords, axis=0)
    coords = coords - shift
    tmp = cart2sph(*coords.T)
    hits = np.zeros(coords.shape[0], dtype=bool)
    r = 2*tmp[0].max()
    dims = dimensionality(network)
    if sum(dims) == 1:
        hi = np.where(coords[:, dims] == coords[:, dims].max())[0]
        lo = np.where(coords[:, dims] == coords[:, dims].min())[0]
        hits[hi] = True
        hits[lo] = True
        return hits
    if sum(dims) == 2:
        markers = generate_points_on_circle(n=max(10, int(coords.shape[0]/10)), r=r)
        pts = np.vstack((coords[:, dims], markers))
    else:
        markers = generate_points_on_sphere(n=max(10, int(coords.shape[0]/10)), r=r)
        pts = np.vstack((coords, markers))
    tri = Delaunay(pts, incremental=False)
    (indices, indptr) = tri.vertex_neighbor_vertices
    for k in range(coords.shape[0], tri.npoints):
        neighbors = indptr[indices[k]:indices[k+1]]
        inds = np.where(neighbors < coords.shape[0])
        hits[neighbors[inds]] = True
    return hits


def internode_distance(network, inds_1=None, inds_2=None):
    r"""
    Find the distance between all nodes on set 1 to each node in set 2

    Parameters
    ----------
    network : dict
        The network dictionary
    inds_1 : array_like
        A list containing the indices of the first set of nodes
    inds_2 : array_Like
        A list containing the indices of the first set of nodes.  It's OK if
        these indices are partially or completely duplicating ``nodes1``.

    Returns
    -------
    dist : array_like
        A distance matrix with ``len(site1)`` rows and ``len(sites2)`` columns.
        The distance between site *i* in ``site1`` and *j* in ``sites2`` is
        located at *(i, j)* and *(j, i)* in the distance matrix.

    Notes
    -----
    This function computes and returns a distance matrix, so can get large.
    For distances between larger sets a KD-tree approach would be better,
    which is available in ``scipy.spatial``.

    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    p1 = np.array(inds_1, ndmin=1)
    p2 = np.array(inds_2, ndmin=1)
    return distance_matrix(coords[p1], coords[p2])


def iscoplanar(network):
    r"""
    Determines if specified nodes are coplanar with each other

    Parameters
    ----------
    network : dict
        The graph dictionary

    Returns
    -------
    flag : bool
        A boolean value of whether given nodes are coplanar (``True``) or
        not (``False``)

    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix + '.coords']
    if np.shape(coords)[0] < 3:
        raise Exception('At least 3 input pores are required')

    Px = coords[:, 0]
    Py = coords[:, 1]
    Pz = coords[:, 2]

    # Do easy check first, for common coordinate
    if np.shape(np.unique(Px))[0] == 1:
        return True
    if np.shape(np.unique(Py))[0] == 1:
        return True
    if np.shape(np.unique(Pz))[0] == 1:
        return True

    # Perform rigorous check using vector algebra
    # Grab first basis vector from list of coords
    n1 = np.array((Px[1] - Px[0], Py[1] - Py[0], Pz[1] - Pz[0])).T
    n = np.array([0.0, 0.0, 0.0])
    i = 1
    while n.sum() == 0:
        if i >= (np.size(Px) - 1):
            return False
        # Chose a secon basis vector
        n2 = np.array((Px[i+1] - Px[i], Py[i+1] - Py[i], Pz[i+1] - Pz[i])).T
        # Find their cross product
        n = np.cross(n1, n2)
        i += 1
    # Create vectors between all other pairs of points
    r = np.array((Px[1:-1] - Px[0], Py[1:-1] - Py[0], Pz[1:-1] - Pz[0]))
    # Ensure they all lie on the same plane
    n_dot = np.dot(n, r)

    return bool(np.sum(np.absolute(n_dot)) == 0)


def is_fully_connected(network, inds=None):
    r"""
    Checks whether graph is fully connected, i.e. not clustered

    Parameters
    ----------
    network : dict
        The network dictionary
    inds : array_like (optional)
        The indices of boundary nodes (i.e. inlets/outlets). If this is given
        the multiple sample spanning clusters will count as fully connected.

    Returns
    -------
    flag : bool
        If ``inds`` is not specified, then returns ``True`` only if
        the entire network is connected to the same cluster. If
        ``inds`` is given, then returns ``True`` only if all clusters
        are connected to the given boundary nodes.
    """
    am = dict_to_am(network)
    am = am.tolil()
    inds = np.array(inds)
    temp = csgraph.connected_components(am, directed=False)[1]
    is_connected = np.unique(temp).size == 1
    Np = am.shape[0]
    Nt = int(am.nnz/2)
    # Ensure all clusters are part of inds, if given
    if not is_connected and inds is not None:
        am.resize(Np + 1, Np + 1)
        am.rows[-1] = inds.tolist()
        am.data[-1] = np.arange(Nt, Nt + len(inds)).tolist()
        temp = csgraph.connected_components(am, directed=False)[1]
        is_connected = np.unique(temp).size == 1
    return is_connected


def get_cubic_spacing(network):
    r"""
    Determine spacing of a cubic network

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    spacing : ndarray
        An array containing the spacing between nodes in each direction

    """
    node_prefix = get_node_prefix(network)
    edge_prefix = get_edge_prefix(network)
    coords = network[node_prefix+'.coords']
    conns = network[edge_prefix+'.conns']
    # Find Network spacing
    C12 = coords[conns]
    mag = np.linalg.norm(np.diff(C12, axis=1), axis=2)
    unit_vec = np.around(np.squeeze(np.diff(C12, axis=1)) / mag, decimals=14)
    spacing = [0, 0, 0]
    dims = dimensionality(network)
    # Ensure vectors point in n-dims unique directions
    c = {tuple(row): 1 for row in unit_vec}
    mag = np.atleast_1d(mag.squeeze()).astype(float)
    if len(c.keys()) > sum(dims):
        raise Exception(
            "Spacing is undefined when throats point in more directions"
            " than network has dimensions."
        )
    for ax in [0, 1, 2]:
        if dims[ax]:
            inds = np.where(unit_vec[:, ax] == unit_vec[:, ax].max())[0]
            temp = np.unique(mag[inds])
            if not np.allclose(temp, temp[0]):
                raise Exception("A unique value of spacing could not be found.")
            spacing[ax] = temp[0]
    return np.array(spacing)


def get_cubic_shape(network):
    r"""
    Determine shape of a cubic network

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    shape : ndarray
        An array containing the shape of the network each direction

    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    L = np.ptp(coords, axis=0)
    mask = L.astype(bool)
    S = get_cubic_spacing(network)
    shape = np.array([1, 1, 1], int)
    shape[mask] = L[mask] / S[mask] + 1
    return shape


def get_domain_area(network, inlets=None, outlets=None):
    r"""
    Determine the cross sectional area relative to the inlets/outlets.

    Parameters
    ----------
    network : dict
        The network dictionary
    inlets : array_like
        The indices of the inlets
    outlets : array_Like
        The indices of the outlets

    Returns
    -------
    area : scalar
        The cross sectional area relative to the inlets/outlets.

    """
    if dimensionality(network).sum() != 3:
        raise Exception('The network is not 3D, specify area manually')
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    inlets = coords[inlets]
    outlets = coords[outlets]
    if not iscoplanar(inlets):
        print('Detected inlet pores are not coplanar')
    if not iscoplanar(outlets):
        print('Detected outlet pores are not coplanar')
    Nin = np.ptp(inlets, axis=0) > 0
    if Nin.all():
        print('Detected inlets are not oriented along a principle axis')
    Nout = np.ptp(outlets, axis=0) > 0
    if Nout.all():
        print('Detected outlets are not oriented along a principle axis')
    hull_in = ConvexHull(points=inlets[:, Nin])
    hull_out = ConvexHull(points=outlets[:, Nout])
    if hull_in.volume != hull_out.volume:
        print('Inlet and outlet faces are different area')
    area = hull_in.volume  # In 2D: volume=area, area=perimeter
    return area


def get_domain_length(network, inlets=None, outlets=None):
    r"""
    Determine the domain length relative to the inlets/outlets.

    Parameters
    ----------
    network : dict
        The network dictionary
    inlets : array_like
        The pore indices of the inlets.
    outlets : array_Like
        The pore indices of the outlets.

    Returns
    -------
    area : scalar
        The domain length relative to the inlets/outlets.

    """
    node_prefix = get_node_prefix(network)
    coords = network[node_prefix+'.coords']
    inlets = coords[inlets]
    outlets = coords[outlets]
    if not iscoplanar(inlets):
        print('Detected inlet pores are not coplanar')
    if not iscoplanar(outlets):
        print('Detected inlet pores are not coplanar')
    tree = KDTree(data=inlets)
    Ls = np.unique(np.float64(tree.query(x=outlets)[0]))
    if not np.allclose(Ls, Ls[0]):
        print('A unique value of length could not be found')
    length = Ls[0]
    return length


def tri_to_am(tri):
    r"""
    Given a Delaunay triangulation object from Scipy's ``spatial`` module,
    converts to a sparse adjacency matrix network representation.

    Parameters
    ----------
    tri : Delaunay Triangulation Object
        This object is produced by ``scipy.spatial.Delaunay``

    Returns
    -------
    A sparse adjacency matrix in COO format.  The network is undirected
    and unweighted, so the adjacency matrix is upper-triangular and all the
    weights are set to 1.

    """
    # Create an empty list-of-list matrix
    lil = sprs.lil_matrix((tri.npoints, tri.npoints))
    # Scan through Delaunay triangulation object to retrieve pairs
    indices, indptr = tri.vertex_neighbor_vertices
    if 1:  # Original way
        for k in range(tri.npoints):
            lil.rows[k] = indptr[indices[k]:indices[k + 1]].tolist()
        # Convert to coo format
        lil.data = lil.rows  # Just a dummy array to make things work properly
        coo = lil.tocoo()
    else:  # Alternative way, about 10x SLOWER
        coo = [[], []]
        for k in range(tri.npoints):
            # lil.rows[k] = indptr[indices[k]:indices[k + 1]].tolist()
            col = indptr[indices[k]:indices[k+1]]
            coo[0].extend(np.ones_like(col)*k)
            coo[1].extend(col)
        coo = sprs.coo_matrix((np.ones_like(coo[0]), (coo[0], coo[1])))
    # Set weights to 1's
    coo.data = np.ones_like(coo.data)
    # Remove diagonal, and convert to csr remove duplicates
    am = sprs.triu(A=coo, k=1, format='csr')
    # The convert back to COO and return
    am = am.tocoo()
    return am


def vor_to_am(vor):
    r"""
    Given a Voronoi tessellation object from Scipy's ``spatial`` module,
    converts to a sparse adjacency matrix network representation in COO format.

    Parameters
    ----------
    vor : Voronoi Tessellation object
        This object is produced by ``scipy.spatial.Voronoi``

    Returns
    -------
    A sparse adjacency matrix in COO format.  The network is undirected
    and unweighted, so the adjacency matrix is upper-triangular and all the
    weights are set to 1.

    """
    if 0:  # Original way, 2X slower
        rc = [[], []]
        for ij in vor.ridge_dict.keys():
            row = vor.ridge_dict[ij].copy()
            # Make sure voronoi cell closes upon itself
            row.append(row[0])
            # Add connections to rc list
            rc[0].extend(row[:-1])
            rc[1].extend(row[1:])
        rc = np.vstack(rc).T
    else:
        v = vor.ridge_vertices.copy()
        # Add row [0] to close the facet on itself, add -1 to break connection to
        # next facet in list as connections with -1 get deleted
        _ = [row.extend([row[0], -1]) for row in v]
        v = np.hstack(v)
        rc = np.vstack((v[:-1], v[1:])).T
    mask = np.any(rc < 0, axis=1)
    rc = rc[~mask]
    rc = np.sort(rc, axis=1)
    am = conns_to_am(rc)
    return am


def dict_to_am(network, weights=None):
    r"""
    Convert a graph dictionary into a ``scipy.sparse`` adjacency matrix in
    COO format

    Parameters
    ----------
    network : dict
        A network dictionary
    weights : ndarray, optional
        The weight values to use for the connections. If not provided
        then 1's are assumed.

    Returns
    -------
    am : sparse matrix
        The sparse adjacency matrix in COO format

    Notes
    -----
    If the edge connections in ``g`` are in upper-triangular form, then the
    graph is assumed to be undirected and the returned adjacency matrix is
    symmetrical (i.e. the triu entries are reflected in tril). If any edges
    are found in the lower triangle, then the returned adjacency matrix is
    unsymmetrical.

    Multigraphs (i.e. duplicate connections between nodes) are not suported,
    but this is not checked for here to avoid overhead since this function is
    called frequently.

    """
    edge_prefix = get_edge_prefix(network)
    node_prefix = get_node_prefix(network)
    conns = np.copy(network[edge_prefix+'.conns'])
    shape = [network[node_prefix+'.coords'].shape[0]]*2
    if weights is None:
        weights = np.ones_like(conns[:, 0], dtype=int)
    if isgtriu(network):  # If graph is triu, then it is assumed to be undirected
        conns = np.vstack((conns, np.fliplr(conns)))  # Reflect to tril
        data = np.ones_like(conns[:, 0], dtype=int)  # Generate fake data
        am = sprs.coo_matrix((data, (conns[:, 0], conns[:, 1])), shape=shape)
        am.data = np.hstack((weights, weights))
    else:
        am = sprs.coo_matrix((weights, (conns[:, 0], conns[:, 1])), shape=shape)
    return am


def dict_to_im(network):
    r"""
    Convert a graph dictionary into a ``scipy.sparse`` incidence matrix in COO
    format

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    im : sparse matrix
        The sparse incidence matrix in COO format

    Notes
    -----
    Rows correspond to nodes and columns correspond to edges. Each column
    has 2 nonzero values indicating which 2 nodes are connected by the
    corresponding edge. Each row contains an arbitrary number of nonzeros
    whose locations indicate which edges are directly connected to the
    corresponding node.
    """
    edge_prefix = get_edge_prefix(network)
    node_prefix = get_node_prefix(network)
    conns = network[edge_prefix+'.conns']
    coords = network[node_prefix+'.coords']
    if isgtriu(network):
        data = np.ones(2*conns.shape[0], dtype=int)
        shape = (coords.shape[0], conns.shape[0])
        temp = np.arange(conns.shape[0])
        cols = np.vstack((temp, temp)).T.flatten()
        rows = conns.flatten()
        im = sprs.coo_matrix((data, (rows, cols)), shape=shape)
    else:
        raise Exception('This function is not implemented for directed graphs')
    return im


def ismultigraph(network):
    r"""
    Checks if graph contains multiple connections between any pair of nodes

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    flag : bool
        Returns ``True`` if any pair of nodes is connected by more than one
        edge.
    """
    edge_prefix = get_edge_prefix(network)
    node_prefix = get_node_prefix(network)
    conns = network[edge_prefix+'.conns']
    coords = network[node_prefix+'.coords']
    data = np.ones_like(conns[:, 0], dtype=int)
    shape = 2*[coords.shape[0]]
    am = sprs.coo_matrix((data, (conns[:, 0], conns[:, 1])), shape=shape)
    am.sum_duplicates()
    return np.any(am.data > 1)


def isgtriu(network):
    r"""
    Determines if graph connections are in upper triangular format

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    flag : bool
        Returns ``True`` if *all* rows in "conns" are ordered as [lo, hi]
    """
    edge_prefix = get_edge_prefix(network)
    conns = network[edge_prefix+'.conns']
    return np.all(conns[:, 0] < conns[:, 1])


def to_triu(network):
    r"""
    Adjusts conns array to force into upper triangular form

    Parameters
    ----------
    network : dict
        The network dictionary

    Returns
    -------
    network : dict
        The graph dictionary with edge connections updated

    Notes
    -----
    This does not check for the creation of duplicate connections
    """
    edge_prefix = get_edge_prefix(network)
    conns = network[edge_prefix+'.conns']
    network[edge_prefix+'.conns'] = np.sort(conns, axis=1)
    return network


def conns_to_am(conns, shape=None, force_triu=True, drop_diag=True,
                drop_dupes=True, drop_negs=True):
    r"""
    Converts a list of connections into a Scipy sparse adjacency matrix

    Parameters
    ----------
    conns : array_like, N x 2
        The list of site-to-site connections
    shape : list, optional
        The shape of the array.  If none is given then it is taken as 1 + the
        maximum value in ``conns``.
    force_triu : bool
        If True (default), then all connections are assumed undirected, and
        moved to the upper triangular portion of the array
    drop_diag : bool
        If True (default), then connections from a site and itself are removed.
    drop_dupes : bool
        If True (default), then all pairs of sites sharing multiple connections
        are reduced to a single connection.
    drop_negs : bool
        If True (default), then all connections with one or both ends pointing
        to a negative number are removed.

    Returns
    -------
    am : ndarray
        A sparse adjacency matrix in COO format

    """
    if drop_negs:  # Remove connections to -1
        keep = ~np.any(conns < 0, axis=1)
        conns = conns[keep]
    if drop_diag:  # Remove connections of [self, self]
        keep = np.where(conns[:, 0] != conns[:, 1])[0]
        conns = conns[keep]
    if force_triu:  # Sort connections to [low, high]
        conns = np.sort(conns, axis=1)
    # Now convert to actual sparse array in COO format
    data = np.ones_like(conns[:, 0], dtype=int)
    if shape is None:
        N = conns.max() + 1
        shape = (N, N)
    am = sprs.coo_matrix((data, (conns[:, 0], conns[:, 1])), shape=shape)
    if drop_dupes:  # Convert to csr and back too coo
        am = am.tocsr()
        am = am.tocoo()
    # Perform one last check on adjacency matrix
    missing = np.where(np.bincount(conns.flatten()) == 0)[0]
    if np.size(missing) or np.any(am.col.max() < (shape[0] - 1)):
        print('Some nodes are not connected to any bonds')
    return am


def istriu(am):
    r"""
    Returns ``True`` if the sparse adjacency matrix is upper triangular
    """
    if am.shape[0] != am.shape[1]:
        print('Matrix is not square, triangularity is irrelevant')
        return False
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    return np.all(am.row <= am.col)


def istril(am):
    r"""
    Returns ``True`` if the sparse adjacency matrix is lower triangular
    """
    if am.shape[0] != am.shape[1]:
        print('Matrix is not square, triangularity is irrelevant')
        return False
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    return np.all(am.row >= am.col)


def istriangular(am):
    r"""
    Returns ``True`` if the sparse adjacency matrix is either upper or lower
    triangular
    """
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    return istril(am) or istriu(am)


def issymmetric(am):
    r"""
    A method to check if a square matrix is symmetric
    Returns ``True`` if the sparse adjacency matrix is symmetric
    """
    if am.shape[0] != am.shape[1]:
        print('Matrix is not square, symmetrical is irrelevant')
        return False
    if am.format != 'coo':
        am = am.tocoo(copy=False)
    if istril(am) or istriu(am):
        return False
    # Compare am with its transpose, element wise
    sym = ((am != am.T).size) == 0
    return sym



from openpnm.models.phase import _phasedocs
import numpy as np


__all__ = [
    "ideal_gas",
    "water_correlation",
    "liquid_mixture_COSTALD",
    "liquid_pure_COSTALD",
    "mass_to_molar",
]


@_phasedocs
def ideal_gas(
    phase,
    P='pore.pressure',
    T='pore.temperature',
    MW='param.molecular_weight',
):
    r"""
    Uses ideal gas law to calculate the mass density of an ideal gas

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(P)s
    %(MW)s

    Returns
    -------

    """
    P = phase[P]
    T = phase[T]
    try:
        # If phase is a pure species, it should have molecular weight in params
        MW = phase[MW]
    except KeyError:
        # Otherwise, get the mole weighted average value
        MW = phase.get_mix_vals(MW)
    R = 8.314462618  # J/(mol.K)
    value = P/(R*T)*(MW/1000)  # Convert to kg/m3
    return value


@_phasedocs
def water_correlation(
    phase,
    T='pore.temperature',
    salinity='pore.salinity',
):
    r"""
    Calculates density of pure water or seawater at atmospheric pressure
    using Eq. (8) given by Sharqawy et. al [1]. Values at temperature higher
    than the normal boiling temperature are calculated at the saturation
    pressure.

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(salinity)s

    Returns
    -------

    Notes
    -----
    ``T`` must be in K, and ``salinity`` in g of salt per kg of phase, or
    ppt (parts per thousand)
    VALIDITY: 273 < T < 453 K; 0 < S < 160 g/kg;
    ACCURACY: 0.1 pct

    References
    ----------
    [1] Sharqawy M. H., Lienhard J. H., and Zubair, S. M., Desalination and
        Water Treatment, 2010.

    """
    T = phase[T]
    if salinity in phase.keys():
        S = phase[salinity]
    else:
        S = 0
    a1 = 9.9992293295E+02
    a2 = 2.0341179217E-02
    a3 = -6.1624591598E-03
    a4 = 2.2614664708E-05
    a5 = -4.6570659168E-08
    b1 = 8.0200240891E-01
    b2 = -2.0005183488E-03
    b3 = 1.6771024982E-05
    b4 = -3.0600536746E-08
    b5 = -1.6132224742E-11
    TC = T-273.15
    rho_w = a1 + a2*TC + a3*TC**2 + a4*TC**3 + a5*TC**4
    d_rho = b1*S + b2*S*TC + b3*S*(TC**2) + b4*S*(TC**3) + b5*(S**2)*(TC**2)
    rho_sw = rho_w + d_rho
    value = rho_sw
    return value


@_phasedocs
def liquid_mixture_COSTALD(
    phase,
    T='pore.temperature',
    MWs='param.molecular_weight.*',
    Tcs='param.critical_temperature.*',
    Vcs='param.critical_volume.*',
    omegas='param.acentric_factor.*',
):
    r"""
    Computes the density of a liquid mixture using the COrrospoding STAtes
    Liquid Density (COSTALD) method.

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(MWs)s
    %(Tcs)s
    %(Vcs)s
    %(omegas)s

    Returns
    -------
    density : ndarray
        The density of the liquid mixture in units of kg/m3.  Note that
        ``chemicals.volume.COSTALD`` returns molar volume, so this function
        converts it to density using the mole fraction weighted molecular
        weight of the mixture: :math:`MW_{mix} = \Sigma x_i \cdot MW_i`.

    Notes
    -----
    This is same approach used by ``chemicals.volume.COSTALD_mixture`` and
    exact numerical correspondance is confirmed. Unlike the ``chemicals``
    version, this function is vectorized over the conditions rather than
    the compositions, since a typical simulation has millions of pores each
    representing an independent conditions, but a mixture typically only has
    a few components.

    """
    # Fetch parameters for each pure component
    Tcs = phase.get_comp_vals(Tcs)
    Vcs = phase.get_comp_vals(Vcs)
    omegas = phase.get_comp_vals(omegas)
    Xs = phase['pore.mole_fraction']
    # Compute mixture values
    omegam = np.vstack([Xs[k]*omegas[k] for k in Xs.keys()]).sum(axis=0)
    Vm1 = np.vstack([Xs[k]*Vcs[k] for k in Xs.keys()]).sum(axis=0)
    Vm2 = np.vstack([Xs[k]*(Vcs[k])**(2/3) for k in Xs.keys()]).sum(axis=0)
    Vm3 = np.vstack([Xs[k]*(Vcs[k])**(1/3) for k in Xs.keys()]).sum(axis=0)
    Vm = 0.25*(Vm1 + 3*Vm2*Vm3)
    Tcm = 0.0
    for i, ki in enumerate(Xs.keys()):
        inner = 0.0
        for j, kj in enumerate(Xs.keys()):
            inner += Xs[ki]*Xs[kj]*(Vcs[ki]*Tcs[ki]*Vcs[kj]*Tcs[kj])**0.5
        Tcm += inner
    Tcm = Tcm/Vm
    # Convert molar volume to normal mass density
    MWs = phase.get_comp_vals('param.molecular_weight')
    MWm = np.vstack([Xs[k]*MWs[k] for k in Xs.keys()]).sum(axis=0)
    T = phase[T]
    rhoL = liquid_pure_COSTALD(
        phase=phase,
        T=T,
        MW=MWm,
        Tc=Tcm,
        Vc=Vm,
        omega=omegam,
    )
    return rhoL


@_phasedocs
def liquid_pure_COSTALD(
    phase,
    T='pore.temperature',
    Tc='param.critical_temperature',
    Vc='param.critical_volume',
    omega='param.acentric_factor',
    MW='param.molecular_weight',
):
    r"""
    Computes the density of a pure liquid using the COrrospoding STAtes
    Liquid Density (COSTALD) method.

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(Tc)s
    %(Vc)s
    %(omega)s
    %(MW)s

    Returns
    -------

    """
    Vc = phase[Vc]
    Tc = phase[Tc]
    omega = phase[omega]
    T = phase[T]
    Tr = T/Tc
    V0 = 1 - 1.52816*(1-Tr)**(1/3) + 1.43907*(1-Tr)**(2/3) - 0.81446*(1-Tr) + \
        0.190454*(1-Tr)**(4/3)
    V1 = (-0.296123 + 0.386914*Tr - 0.0427258*Tr**2 - 0.0480645*Tr**3)/(Tr - 1.00001)
    Vs = Vc*V0*(1-omega*V1)
    MW = phase[MW]
    rhoL = 1e-3*MW/Vs
    return rhoL


@_phasedocs
def mass_to_molar(
    phase,
    MW='param.molecular_weight',
    rho='pore.density',
):
    r"""
    Calculates the molar density from the molecular weight and mass density

    Parameters
    ----------
    %(phase)s
    %(MW)s
    %(rho)s

    Returns
    -------
    value : ndarray
        A numpy ndrray containing molar density values [mol/m3]

    """
    MW = phase[MW]/1000
    rho = phase[rho]
    value = rho/MW
    return value



r"""
Utilities and helper classes/functions
======================================

This module contains two very important classes (Project and Workspace)
as well as a number of helper classes.

"""

from re import L
from ._misc import *
from ._settings import *
from ._workspace import *
from ._project import *
from ._health import *


def _get_version():
    from openpnm.__version__ import __version__ as ver

    suffix = ".dev0"
    if ver.endswith(suffix):
        ver = ver[: -len(suffix)]
    return ver


def _setup_logger_rich():
    import logging
    from rich.logging import RichHandler

    FORMAT = "%(message)s"
    logging.basicConfig(
        format=FORMAT, datefmt="[%X]", handlers=[RichHandler(rich_tracebacks=True)]
    )


_setup_logger_rich()



import scipy.spatial as sptl
import scipy.sparse as sprs
import numpy as np
from openpnm._skgraph.generators import tools
from openpnm._skgraph.operations import trim_nodes
from openpnm._skgraph.tools import isoutside, conns_to_am
from openpnm._skgraph.queries import find_neighbor_nodes


def voronoi_delaunay_dual(
    points,
    shape,
    trim=True,
    reflect=True,
    f=1,
    relaxation=0,
    node_prefix='node',
    edge_prefix='edge',
    return_tri=False,
):
    r"""
    Generate a dual Voronoi-Delaunay network from given base points

    Parameters
    ----------
    points : array_like or scalar
        The points to be tessellated.  If a scalar is given a set of points
        of that size is generated inside the given ``shape``.
    shape : array_like
        The size of the domain in which the points lie
    trim : bool, optional
        If ``True`` (default) then all points lying beyond the given domain
        shape will be removed
    reflect : bool, optionl
        If ``True`` (Default) then points are reflected across each face of the
        domain prior to performing the tessellation.
    f : float
        The fraction of points which should be reflected.  The default is 1 which
        reflects all the points in the domain, but this can lead to a lot of
        unnecessary points, so setting to 0.1 or 0.2 helps speed, but risks that
        the tessellation may not have smooth faces if not enough points are
        reflected.
    relaxation : int, optional (default = 0)
        The number of iterations to use for relaxing the base points. This is
        sometimes called `Lloyd's algorithm
        <https://en.wikipedia.org/wiki/Lloyd%27s_algorithm>`_. This function computes
        the new base points as the simple average of the Voronoi vertices instead
        of rigorously finding the center of mass, which is quite time consuming.
        To use the rigorous method, call the ``lloyd_relaxation`` function manually
        to obtain relaxed points, then pass the points directly to this funcion.
        The results are quite stable after only a few iterations.

    Returns
    -------
    network : dict
        A dictionary containing '<node_prefix>.coords' and '<edge_prefix>.conns'
    vor : Voronoi object
        The Voronoi tessellation object produced by ``scipy.spatial.Voronoi``
    tri : Delaunay object
        The Delaunay triangulation object produced by ``scipy.spatial.Delaunay``

    """
    # Generate a set of base points if scalar was given
    points = tools.parse_points(
        points=points,
        shape=shape,
        reflect=reflect,
        f=f,
    )

    # Generate mask to remove any dims with all 0's
    mask = ~np.all(points == 0, axis=0)

    # Perform tessellations
    vor = sptl.Voronoi(points=points[:, mask])
    for _ in range(relaxation):
        points = tools.lloyd_relaxation(vor, mode='fast')
        vor = sptl.Voronoi(points=points[:, mask])

    # Collect delaunay edges
    conns_del = vor.ridge_points
    # Deal with voronoi edges
    v = vor.ridge_vertices.copy()  # Assuming 'ridge' means facet between regions
    # Add row [0] to close the facet on itself, add -1 to break connection to
    # next facet in list as connections with -1 get deleted later
    _ = [row.extend([row[0], -1]) for row in v]
    v = np.hstack(v)
    conns_vor = np.vstack((v[:-1], v[1:])).T
    mask = np.any(conns_vor < 0, axis=1)
    conns_vor = conns_vor[~mask] + vor.npoints
    # Finally, get interconnecting edges
    idx = [vor.regions[vor.point_region[i]] for i in range(0, len(vor.regions)-1)]
    conns_inter = [([i]*len(idx[i]), idx[i]) for i in range(0, len(idx))]
    conns_inter = np.hstack(conns_inter).astype(int).T
    mask = np.any(conns_inter < 0, axis=1)
    conns_inter = conns_inter[~mask, :] + np.array([0, vor.npoints], dtype=int)
    conns = np.vstack((conns_del, conns_vor, conns_inter))

    # Tidy up
    am = conns_to_am(conns)
    conns = np.vstack((am.row, am.col)).T

    # Combine delaunay and voronoi points
    pts_all = np.vstack((vor.points, vor.vertices))
    # Rounding is crucial since some voronoi verts endup outside domain
    pts_all = np.around(pts_all, decimals=10)
    # Convert coords to 3D if necessary
    mask = ~np.all(points == 0, axis=0)
    if mask.sum() < 3:
        coords = np.zeros([pts_all.shape[0], 3], dtype=float)
        coords[:, mask] = pts_all
    else:
        coords = pts_all

    # Assign coords and conns to network dict
    network = {}
    network[node_prefix+'.coords'] = coords
    network[edge_prefix+'.conns'] = conns

    n_nodes = coords.shape[0]
    n_edges = conns.shape[0]

    # Label all pores and throats by type
    network[node_prefix+'.delaunay'] = np.zeros(n_nodes, dtype=bool)
    network[node_prefix+'.delaunay'][0:vor.npoints] = True
    network[node_prefix+'.voronoi'] = np.zeros(n_nodes, dtype=bool)
    network[node_prefix+'.voronoi'][vor.npoints:] = True
    # Label throats between Delaunay pores
    network[edge_prefix+'.delaunay'] = np.zeros(n_edges, dtype=bool)
    Ts = np.all(network[edge_prefix+'.conns'] < vor.npoints, axis=1)
    network[edge_prefix+'.delaunay'][Ts] = True
    # Label throats between Voronoi pores
    network[edge_prefix+'.voronoi'] = np.zeros(n_edges, dtype=bool)
    Ts = np.all(network[edge_prefix+'.conns'] >= vor.npoints, axis=1)
    network[edge_prefix+'.voronoi'][Ts] = True
    # Label throats connecting a Delaunay and a Voronoi pore
    Ts = np.sum(network[node_prefix+'.delaunay'][conns].astype(int), axis=1) == 1
    network[edge_prefix+'.interconnect'] = Ts

    if trim:
        # Find all delaunay nodes outside the domain
        Ps = isoutside(network=network, shape=shape)*network[node_prefix+'.delaunay']
        if np.any(Ps):  # only occurs if points were reflected
            # Find voronoi nodes connected to these and mark them as surface nodes
            inds = np.where(Ps)[0]
            Ns = find_neighbor_nodes(network=network, inds=inds)
            network[node_prefix+'.surface'] = np.zeros(n_nodes, dtype=bool)
            network[node_prefix+'.surface'][Ns] = True
            Ps = isoutside(network=network, shape=shape)
            inds = np.where(Ps)[0]
            network = trim_nodes(network=network, inds=inds)
        else:
            trim = isoutside(network=network, shape=shape)
            inds = np.where(trim)[0]
            network = trim_nodes(network=network, inds=inds)

    if return_tri:
        tri = sptl.Delaunay(points=points[:, mask])
    else:
        tri = None
    return network, vor, tri


if __name__ == "__main__":
    import openpnm as op
    pn, vor, tri = voronoi_delaunay_dual(
        points=500,
        shape=[1, 1, 0],
        trim=True,
        reflect=True,
        f=0.2,
        relaxation=0,
        node_prefix='pore',
        edge_prefix='throat',
    )
    net = op.network.Network()
    net.update(pn)
    h = None
    h = op.visualization.plot_connections(
        net, throats=pn['throat.delaunay'], c='b', ax=h)
    h = op.visualization.plot_connections(
        net, throats=pn['throat.voronoi'], c='g', ax=h)
    h = op.visualization.plot_connections(
        net, throats=pn['throat.interconnect'], c='r', ax=h)
    h = op.visualization.plot_coordinates(
        net, pores=pn['pore.voronoi'], c='g', markersize=150, ax=h)
    h = op.visualization.plot_coordinates(
        net, pores=pn['pore.delaunay'], c='b', markersize=150, ax=h)



import logging
import inspect
import openpnm as op
import numpy as np
from copy import deepcopy
from openpnm.utils import (
    PrintableDict,
    Workspace,
    is_valid_propname,
)


logger = logging.getLogger(__name__)
ws = Workspace()


__all__ = [
    'ModelsDict',
    'ModelWrapper',
    'ModelsMixin2',
]


class ModelsDict(PrintableDict):
    """
    This subclassed dictionary is assigned to the ``models`` attribute of
    all objects that inherit from the ``ModelsMixin`` class.  Each
    dictionary entry corresponds to an entry in the target object's
    dictionary, and contains the models and associated parameters for
    generating the model.

    The main features of this subclass are three methods the help resolve
    the order in which models should be called: ``dependency_list``,
    ``dependency_graph``, and ``dependency_map``.

    """

    def _find_target(self):
        """
        Finds and returns the target object to which this ModelsDict is
        associated.
        """
        for proj in ws.values():
            for obj in proj:
                if hasattr(obj, "models"):
                    if obj.models is self:
                        return obj
        raise Exception("No target object found!")

    def dependency_list(self):
        r"""
        Returns a list of dependencies in the order with which they should
        be called to ensure data is calculated by one model before it's
        asked for by another.

        Notes
        -----
        This raises an exception if the graph has cycles which means the
        dependencies are unresolvable (i.e. there is no order which the
        models can be called that will work).  In this case it is possible
        to visually inspect the graph using ``dependency_graph``.

        See Also
        --------
        dependency_graph
        dependency_map

        """
        import networkx as nx

        dtree = self.dependency_graph()
        cycles = list(nx.simple_cycles(dtree))
        if cycles:
            msg = 'Cyclic dependency: ' + ' -> '.join(cycles[0] + [cycles[0][0]])
            raise Exception(msg)
        d = nx.algorithms.dag.lexicographical_topological_sort(dtree, sorted)
        return list(d)

    def dependency_graph(self, deep=False):
        """
        Returns a NetworkX graph object of the dependencies

        Parameters
        ----------
        deep : bool, optional
            Defines whether intra- or inter-object dependency graph is
            desired. Default is False, i.e. only returns dependencies
            within the object.

        See Also
        --------
        dependency_list
        dependency_map

        """
        import networkx as nx

        dtree = nx.DiGraph()
        models = list(self.keys())

        for model in models:
            propname = model.split("@")[0]
            dtree.add_node(propname)
            # Filter pore/throat props only
            args = op.utils.flat_list(self[model].values())
            dependencies = [arg for arg in args if is_valid_propname(arg)]
            # Add dependency from model's parameters
            for d in dependencies:
                dtree.add_edge(d, propname)

        return dtree

    def dependency_map(self,
                       ax=None,
                       figsize=None,
                       deep=False,
                       style='shell'):  # pragma: no cover
        """
        Create a graph of the dependency graph in a decent format

        Parameters
        ----------
        ax : matplotlib.axis, optional
            Matplotlib axis object on which dependency map is to be drawn.
        figsize : tuple, optional
            Tuple containing frame size.

        See Also
        --------
        dependency_graph
        dependency_list

        """
        import networkx as nx
        import matplotlib.pyplot as plt

        if ax is None:
            fig, ax = plt.subplots()
        if figsize is not None:
            fig.set_size_inches(figsize)

        labels = {}
        node_shapes = {}
        dtree = self.dependency_graph(deep=deep)
        for node in dtree.nodes:
            labels[node] = node.split(".")[1]
            node_shapes[node] = "o" if node.startswith("pore") else "s"
        nx.set_node_attributes(dtree, node_shapes, "node_shape")

        layout = getattr(nx, f"{style}_layout")
        pos = layout(dtree)

        Pprops = [prop for prop in dtree.nodes if prop.startswith("pore")]
        Tprops = [prop for prop in dtree.nodes if prop.startswith("throat")]
        colors = ["yellowgreen", "coral"]
        shapes = ["o", "s"]

        for props, color, shape in zip([Pprops, Tprops], colors, shapes):
            nx.draw(
                dtree,
                pos=pos,
                nodelist=props,
                node_shape=shape,
                labels=labels,
                with_labels=True,
                edge_color='lightgrey',
                node_color=color,
                font_size=12,
                width=2.0
            )

        ax = plt.gca()
        ax.margins(x=0.2, y=0.05)

        return ax

    @property
    def _info(self):  # Pragma: no cover
        r"""
        Prints a nicely formatted list of model names and the domain to which
        they apply.

        Notes
        -----
        This is a hidden function for now, but could be exposed if useful.
        """
        names = {}
        for item in self:
            name, _, domain = item.partition('@')
            if name not in names.keys():
                names[name] = []
            names[name].append(domain)
        D = PrintableDict(names, key='Model', value='Domain')
        print(D)

    def __str__(self):  # pragma: no cover
        horizontal_rule = '―' * 85
        lines = [horizontal_rule]
        strg = '{0:<3s} {1:<35s} {2:<25s} {3}'
        lines.append(strg.format('#', 'Property Name', 'Parameter', 'Value'))
        lines.append(horizontal_rule)
        for i, item in enumerate(self.keys()):
            temp = self[item].copy()
            regen_mode = temp.pop('regen_mode', None)
            model = str(temp.pop('model')).split(' ')[1]
            lines.append(strg.format(str(i+1), item, 'model:', model))
            for param in temp.keys():
                lines.append(strg.format('', '', param+':', temp[param]))
            lines.append(strg.format('', '', 'regeneration mode:', regen_mode))
            lines.append(horizontal_rule)
        return '\n'.join(lines)

    def __delitem__(self, key):
        if '@' in key:
            super().__delitem__(key)
        else:  # Delete all models with the same prefix
            for item in list(self.keys()):
                if item.startswith(key):
                    super().__delitem__(item)

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            d = PrintableDict(key='Model', value='Args')
            for k, v in self.items():
                if k.startswith(key+'@'):
                    d[k] = v
            if len(d) > 0:
                return d
            else:
                raise KeyError(key)

    def update(self, d, domain=None):
        # Catch un-run function
        if hasattr(d, '__call__'):
            raise Exception('Received dict argument is a function, try running it')
        parent = self._find_target()
        for k, v in d.items():
            parent.add_model(propname=k, domain=domain, **v)


class ModelWrapper(dict):
    """
    This class is used to hold individual models and provide some extra
    functionality, such as pretty-printing and the ability to run itself.
    """

    def __call__(self):
        model = self['model']
        kwargs = {}
        for k, v in self.items():
            if k not in ['model', 'regen_mode']:
                kwargs[k] = v
        return model(self.target, **kwargs)

    @property
    def name(self):
        for proj in ws.values():
            for obj in proj:
                if hasattr(obj, 'models'):
                    for key, mod in obj.models.items():
                        if mod is self:
                            return key

    @property
    def propname(self):
        return self.name.split('@')[0]

    @property
    def domain(self):
        element, prop = self.name.split('.', 1)
        prop, domain = prop.split('@')
        return element + '.' + domain

    def __str__(self):  # pragma: no cover
        horizontal_rule = '―' * 78
        lines = [horizontal_rule]
        strg = '{0:<25s} {1:<25s} {2}'
        lines.append(strg.format('Property Name', 'Parameter', 'Value'))
        lines.append(horizontal_rule)
        temp = self.copy()
        regen_mode = temp.pop('regen_mode', None)
        model = str(temp.pop('model')).split(' ')[1]
        lines.append(strg.format(self.propname, 'model:', model))
        for param in temp.keys():
            lines.append(strg.format('', param+':', temp[param]))
        lines.append(strg.format('', 'regeneration mode:', regen_mode))
        lines.append(horizontal_rule)
        return '\n'.join(lines)

    @property
    def target(self):
        """
        Finds and returns the object to which this model is assigned
        """
        for proj in ws.values():
            for obj in proj:
                if hasattr(obj, "models"):
                    for mod in obj.models.values():
                        if mod is self:
                            return obj
        raise Exception("No target object found!")


class ModelsMixin2:
    r"""
    This class is added to ``Network`` and ``Phase`` objects under the
    ``models`` attribute. It provides the functionality for storing and
    running pore-scale models.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.models = ModelsDict()

    def add_model(self, propname, model, domain='all', regen_mode='normal',
                  **kwargs):
        r"""
        Add a pore-scale model to the object, along with the desired arguments

        Parameters
        ----------
        propname : str
            The name of the property being computed. E.g. if
            ``propname='pore.diameter'`` then the computed results will be stored
            in ``obj['pore.diameter']``.
        model : function handle
            The function that produces the values
        domain : str
            The label indicating the locations in which results generated by
            ``model`` should be stored. See `Notes` for more details.
        regen_mode : str
            How the model should be regenerated. Options are:

            ============ =====================================================
            regen_mode   description
            ============ =====================================================
            normal       (default) The model is run immediately upon being
                         added, and is also run each time ``regenerate_models``
                         is called.
            deferred     The model is NOT run when added, but is run each time
                         ``regenerate_models`` is called. This is useful for
                         models that depend on other data that may not exist
                         yet.
            constant     The model is run immediately upon being added, but is
                         is not run when ``regenerate_models`` is called,
                         effectively turning the property into a constant.
            ============ =====================================================

        kwargs : keyword arguments
            All additional keyword arguments are passed on to the model

        Notes
        -----
        The ``domain`` argument dictates where the results of ``model`` should
        be stored. For instance, given ``propname='pore.diameter'`` and
        ``domain='left'`` then when `model` is run, the results are stored in
        in the pores labelled left. Note that if ``model`` returns ``Np``
        values, then values not belonging to ``'pore.left'`` are discarded.
        The following steps outline the process:

        1. Find the pore indices:

        .. code-block:: python

          Ps = obj.pores('left')

        2. Run the model:

        .. code-block:: python

          vals = model(**kwargs)

        3. If the model returns a full Np-length array, then extract the
        correct values and apply them to the corresponding locations:

        .. code-block:: python

          if len(vals) == obj.Np:
              obj['pore.diameter'][Ps] = vals[Ps]

        4. If the model was designed to return only the subset of values then:

        .. code-block:: python

          if len(vals) == obj.num_pores('left'):
              obj['pore.diameter'][Ps] = vals

        """
        if '@' in propname:
            propname, domain = propname.split('@')
        elif domain is None:
            domain = self.settings['default_domain']
        element, prop = propname.split('.', 1)
        domain = domain.split('.', 1)[-1]

        # Add model and regen_mode to kwargs dictionary
        kwargs.update({'model': model, 'regen_mode': regen_mode})
        # Insepct model to extract arguments and default values
        kwargs.update(self._inspect_model(model, kwargs))
        self.models[propname+'@'+domain] = ModelWrapper(**kwargs)
        if regen_mode != 'deferred':
            self.run_model(propname+'@'+domain)

    def _inspect_model(self, model, kwargs={}):
        if model.__defaults__:
            vals = list(inspect.getfullargspec(model).defaults)
            keys = inspect.getfullargspec(model).args[-len(vals):]
            for k, v in zip(keys, vals):  # Put defaults into kwargs
                if k not in kwargs:  # Skip if argument was given in kwargs
                    kwargs.update({k: v})
        return kwargs

    def add_model_collection(self, models, domain='all', regen_mode='deferred'):
        r"""
        Add a ``collection`` of several models at once

        Parameters
        ----------
        models : dict
            The collection of models to add.
        regen_mode : str
            By default the models are not regenerated upon addition. See the
            docstring for ``add_model`` for more information.
        domain : str
            The label indicating which locations the supplied collection
            of models should be applied to.

        Notes
        -----
        Collections are dictionaries that are formatted the same as
        ``obj.models``. Several model collections are available in
        ``openpnm.models.collections``.
        """
        models = deepcopy(models)
        for k, v in models.items():
            if 'domain' not in v.keys():
                v['domain'] = domain
            if 'regen_mode' not in v.keys():
                v['regen_mode'] = regen_mode
            self.add_model(propname=k, **v)

    def regenerate_models(self, propnames=None, exclude=[]):
        r"""
        Runs all the models stored in the object's ``models`` attribute

        Parameters
        ----------
        propnames : list of strings
            If given then only the specified models are run
        exclude : list of strings
            If given then these models will *not* be run

        Notes
        -----
        This function will ensure that models are called in the correct order
        such that 'pore.diameter' will be run before 'pore.volume', since
        the diameter is required to compute the volume.
        """
        all_models = self.models.dependency_list()
        # Regenerate all properties by default
        if propnames is None:
            propnames = all_models
        else:
            propnames = np.atleast_1d(propnames).tolist()
        # Remove any that are specifically excluded
        propnames = np.setdiff1d(propnames, exclude).tolist()
        # Reorder given propnames according to dependency tree
        tmp = [e.split("@")[0] for e in propnames]
        idx_sorted = [all_models.index(e) for e in tmp]
        propnames = [elem for i, elem in sorted(zip(idx_sorted, propnames))]
        # Now run each on in sequence
        for item in propnames:
            try:
                self.run_model(item)
            except KeyError as e:
                msg = (f"{item} was not run since the following property"
                       f" is missing: {e}")
                logger.warning(msg)
                self.models[item]['regen_mode'] = 'deferred'

    def run_model(self, propname, domain=None):
        r"""
        Runs the requested model and places the result into the correct
        locations

        Parameters
        ----------
        propname : str
            The name of the model to run.
        domain : str
            The label of the domain for which the model should be run. Passing
            ``propname='pore.diameter@domain1`` and ``domain=None`` is
            equivalent to passing ``propname='pore.diameter`` and
            ``domain=domain1``. Passing ``domain=None`` will regenerate
            all models starting with ``propname``.
        """
        if domain is None:
            if '@' in propname:  # Get domain from propname if present
                propname, _, domain = propname.partition('@')
                self.run_model(propname=propname, domain=domain)
            else:  # No domain means run model for ALL domains
                for item in self.models.keys():
                    if item.startswith(propname+"@"):
                        _, _, domain = item.partition("@")
                        self.run_model(propname=propname, domain=domain)
        else:  # domain was given explicitly
            domain = domain.split('.', 1)[-1]
            element, prop = propname.split('@')[0].split('.', 1)
            propname = f'{element}.{prop}'
            mod_dict = self.models[propname+'@'+domain]
            # Collect kwargs
            kwargs = {'domain': f'{element}.{domain}'}
            for item in mod_dict.keys():
                if item not in ['model', 'regen_mode']:
                    kwargs[item] = mod_dict[item]
            # Deal with models that don't have domain argument yet
            if 'domain' not in inspect.getfullargspec(mod_dict['model']).args:
                _ = kwargs.pop('domain', None)
                vals = mod_dict['model'](self, **kwargs)
                if isinstance(vals, dict):  # Handle models that return a dict
                    for k, v in vals.items():
                        v = np.atleast_1d(v)
                        if v.shape[0] == 1:  # Returned item was a scalar
                            v = np.tile(v, self._count(element))
                        vals[k] = v[self[f'{element}.{domain}']]
                elif isinstance(vals, (int, float)):  # Handle models that return a float
                    vals = np.atleast_1d(vals)
                else:  # Index into full domain result for use below
                    vals = vals[self[f'{element}.{domain}']]
            else:  # Model that accepts domain arg
                vals = mod_dict['model'](self, **kwargs)
            # Finally add model results to self
            if isinstance(vals, np.ndarray):  # If model returns single array
                if propname not in self.keys():
                    temp = self._initialize_empty_array_like(vals, element)
                    self[f'{element}.{prop}'] = temp
                self[propname][self[f'{element}.{domain}']] = vals
            elif isinstance(vals, dict):  # If model returns a dict of arrays
                for k, v in vals.items():
                    if f'{propname}.{k}' not in self.keys():
                        temp = self._initialize_empty_array_like(v, element)
                        self[f'{propname}.{k}'] = temp
                    self[f'{propname}.{k}'][self[f'{element}.{domain}']] = v



import logging
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import Docorator


docstr = Docorator()
logger = logging.getLogger(__name__)


__all__ = ['FickianDiffusion']


@docstr.dedent
class FickianDiffusionSettings():
    r"""

    Parameters
    ----------
    %(ReactiveTransportSettings.parameters)s

    """
    quantity = 'pore.concentration'
    conductance = 'throat.diffusive_conductance'


@docstr.dedent
class FickianDiffusion(ReactiveTransport):
    r"""
    A class to simulate binary diffusion with reactions

    Parameters
    ----------
    %(ReactiveTransport.parameters)s

    Notes
    -----
    Fickian diffusion in porous materials occurs in the void space, but
    becuase the diffusion is confined to pores it is impacted by the porosity
    and tortuosity of the network.  Thus the total diffusive flux through the
    network is reduced.  This class can be used to simualte diffusion-reaction
    in domains with specified boundary conditions, or it can be used
    to calculate the effective diffusivity of the network by applying
    controlled boundary conditions on opposing faces, calculating the
    diffusion rate, and inverting Fick's first law as follows:

    .. math::

        D_{eff} = N_{A}*L/(A*\Delta C_{A})

    """

    def __init__(self, name='fick_?', **kwargs):
        super().__init__(name=name, **kwargs)
        self.settings._update(FickianDiffusionSettings())
        self.name = name



import logging
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import Docorator


__all__ = ['StokesFlow']


logger = logging.getLogger(__name__)
docstr = Docorator()


@docstr.dedent
class StokesFlowSettings:
    r'''

    Parameters
    ----------
    %(ReactiveTransportSettings.parameters)s
    '''
    quantity = 'pore.pressure'
    conductance = 'throat.hydraulic_conductance'


class StokesFlow(ReactiveTransport):
    r"""
    A subclass of LinearTransport to simulate viscous flow.
    """

    def __init__(self, name='stokes_?', **kwargs):
        super().__init__(name=name, **kwargs)
        self.settings._update(StokesFlowSettings())



import openpnm.models.physics as mods


standard = {
    'throat.hydraulic_conductance': {
        'model': mods.hydraulic_conductance.generic_hydraulic,
    },
    'throat.diffusive_conductance': {
        'model': mods.diffusive_conductance.generic_diffusive,
    },
    'throat.entry_pressure': {
        'model': mods.capillary_pressure.washburn,
    },
    'throat.thermal_conductance': {
        'model': mods.thermal_conductance.generic_thermal,
    },
    'throat.electrical_conductance': {
        'model': mods.electrical_conductance.generic_electrical,
    },
    'throat.ad_dif_conductance': {
        'model': mods.ad_dif_conductance.ad_dif,
    },
}



r"""
=======
OpenPNM
=======

OpenPNM is a package for performing pore network simulations of transport in
porous materials.

"""

import logging

from rich.logging import RichHandler

FORMAT = "%(message)s"
logging.basicConfig(
    format=FORMAT, datefmt="[%X]", handlers=[RichHandler(rich_tracebacks=True)]
)

import numpy as _np

from . import (
    _skgraph,
    algorithms,
    contrib,
    core,
    integrators,
    io,
    models,
    network,
    phase,
    solvers,
    topotools,
    utils,
    visualization,
)
from .utils import Project, Workspace

_np.seterr(divide='ignore', invalid='ignore')

__version__ = utils._get_version()
